#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
sys.path.insert(0, '.')
import pickletools
from multiprocessing import Pool, cpu_count
from collections import defaultdict
from bisect import bisect_left
from array import array
from itertools import chain
from math import sin, cos, pi
from contextlib import suppress

from pybiklib.debug import DEBUG_MSG, DEBUG_INDEXONLY, DEBUG_MODELFAST, DEBUG_MAXSIZE5
from pybiklib.utils import epsilon, filebyteorder, get_texcoords_range
# executable script needs absolute imports
from buildlib.geom import roundeps, Vector
from buildlib.modeldef import modeldefs


minsize = 1
maxsize = 10
if DEBUG_MAXSIZE5:  maxsize = 5

dumps = None

def make_dumps(reproducible):
    global dumps
    import io
    import pickle
    if not reproducible:
        dumps = pickle.dumps
        return
    try:
        class _Pickler (pickle._Pickler):
            def save_dict(self, obj):
                if self.bin:
                    self.write(pickle.EMPTY_DICT)
                else:   # proto 0 -- can't use EMPTY_DICT
                    self.write(pickle.MARK + pickle.DICT)
                self.memoize(obj)
                self._batch_setitems(sorted(obj.items(), key=lambda v: str(v)))
                
            pickle._Pickler.dispatch[dict] = save_dict
            
        def _dumps(obj, protocol=None, *, fix_imports=True):
            f = io.BytesIO()
            try:
                _Pickler(f, protocol, fix_imports=fix_imports).dump(obj)
                return f.getvalue()
            except Exception:
                print('warning: using undocumented interface in module pickle failed, models will not be reproducible:')
                sys.excepthook(*sys.exc_info())
                return pickle.dumps(obj, protocol, fix_imports=fix_imports)
        _dumps({1:2, 3:4})
    except Exception:
        print('warning: using undocumented interface in module pickle failed, models will not be reproducible:')
        sys.excepthook(*sys.exc_info())
        _dumps = pickle.dumps
    dumps = _dumps
        
def equal_vector_fuzzy(v1, v2):
    for v1k, v2k in zip(v1, v2):
        if abs(v1k - v2k) > epsilon:
            return False
    return True
    
    
class VectorData:
    __slots__ = 'vectors origs indices sorted cnts dups toindex'.split()
    
    def __init__(self):
        self.vectors = []
        self.origs = {}
        self.indices = []
        self.sorted = []
        self.cnts = 0
        self.dups = 0
        self.toindex = self.toindex_fast if DEBUG_MODELFAST else self.toindex_dedup
        
    def toindex_dedup(self, ovector):
        with suppress(KeyError):
            return self.origs[ovector]
        vector = ovector.rounded()
        vector = [(0. if v == 0. else v) for v in vector]  # 0.0 == -0.0
        self.cnts += 1
        si = bisect_left(self.sorted, vector)
        i = len(self.vectors)
        if si < i and self.sorted[si] == vector:
            self.dups += 1
            i = self.indices[si]
        else:
            self.vectors.append(vector)
            self.sorted.insert(si, vector)
            self.indices.insert(si, i)
        self.origs[ovector] = i
        return i
        
    def toindex_fast(self, ovector):
        with suppress(KeyError):
            return self.origs[ovector]
        vector = ovector.rounded()
        i = len(self.vectors)
        self.vectors.append(vector)
        self.origs[ovector] = i
        return i
        
        
class ModelFactory:
    mtype_attributes = (
            'name', 'mformat', 'sizenames', 'defaultsize',
            'symmetries', 'axes', 'symbols', 'symbolsI', 'faces', 'facekeys',
            'normal_rotation_symbols', 'rotation_symbols', 'rotation_matrices',
            'face_permutations', 'default_rotation', 'reversepick', 'slicesmode')
        
    def __init__(self, modeldef):
        self.modeldef = modeldef()
        self.create_rotations()
        
    def __getattr__(self, attrname):
        return getattr(self.modeldef, attrname)
        
    def _getattr(self, attrname):
        value = getattr(self, attrname)
        if attrname == 'axes':
            return tuple(tuple(v) for v in value)
        return value
        
    def getattrs(self):
        return {attr: self._getattr(attr) for attr in self.mtype_attributes}
        
    @staticmethod
    def _matrix_equal(m1, m2):
        for line1, line2 in zip(m1, m2):
            for value1, value2 in zip(line1, line2):
                if abs(value1 - value2) > epsilon:
                    return False
        return True
        
    @staticmethod
    def _mult_matrix_vector3(matrix, vector):
        return Vector(sum(matrix[i][k]*vector[k] for k in range(3)) for i in range(3))
        
    @staticmethod
    def _mult_matrix(matrix1, matrix2):
        return [[sum(matrix1[i][k]*matrix2[k][j] for k in range(4)) for j in range(4)] for i in range(4)]
        
    @staticmethod
    def _create_rotation(axis, angle):
        angle = angle / 180. * pi
        sa = sin(angle)
        ca = cos(angle)
        e_ca = 1 - ca
        n1 = axis[0]
        n2 = axis[1]
        n3 = axis[2]
        m = [
            [n1*n1*e_ca + ca,    n1*n2*e_ca - n3*sa, n1*n3*e_ca + n2*sa, 0.],
            [n2*n1*e_ca + n3*sa, n2*n2*e_ca + ca,    n2*n3*e_ca - n1*sa, 0.],
            [n3*n1*e_ca - n2*sa, n3*n2*e_ca + n1*sa, n3*n3*e_ca + ca,    0.],
            [0.,    0.,     0.,     1.],
        ]
        #XXX: try to keep the matrix clean
        for y, line in enumerate(m):
            for x, value in enumerate(line):
                if abs(value) < epsilon:
                    m[y][x] = 0.
        return m
        
    def _create_permutation(self, matrix):
        permutation = {}
        for sym1, symI1, axis1 in zip(self.symbols, self.symbolsI, self.axes):
            for sym2, symI2, axis2 in zip(self.symbols, self.symbolsI, self.axes):
                axisR = self._mult_matrix_vector3(matrix, axis1)
                if axisR.equalfuzzy(axis2):
                    permutation[sym2] = sym1
                    permutation[symI2] = symI1
                elif axisR.inversfuzzy(axis2):
                    permutation[symI2] = sym1
                    permutation[sym2] = symI1
        return permutation
        
    def create_rotations(self):
        prim = []
        for axis, sym, symI, symmetry in zip(self.axes, self.symbols, self.symbolsI, self.symmetries):
            angle = 360. / symmetry
            # for some models (towers, bricks) rotations are equal to the invers rotations,
            # but the symbols are different
            prim.append((sym, self._create_rotation(axis, angle)))
            prim.append((symI, self._create_rotation(axis, -angle)))
        self.normal_rotation_symbols = {'': ''}
        transform = [['', [[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]]]]
        for sp, p in prim:
            transform.append([sp, p])
        for sm, m in transform:
            for sp, p in prim:
                n = self._mult_matrix(m, p)
                sn = sm + sp
                for st, t in transform:
                    if self._matrix_equal(t, n):
                        self.normal_rotation_symbols[sn] = st
                        break
                else:
                    self.normal_rotation_symbols[sn] = sn
                    transform.append([sn, n])
        self.rotation_symbols = [s for s, m in transform]
        self.rotation_matrices = [m for s, m in transform]
        self.face_permutations = {s: self._create_permutation(m) for s, m in transform}
        
    def get_rotated_position(self, cells):
        rotated_position = {}
        centers = [c.center() for c in cells]
        for b, center in enumerate(centers):
            for sym, rotation in zip(self.rotation_symbols, self.rotation_matrices):
                coords = self._mult_matrix_vector3(rotation, center)
                for p, center2 in enumerate(centers):
                    if equal_vector_fuzzy(center2, coords):
                        if sym not in rotated_position:
                            rotated_position[sym] = [0] * len(cells)
                        rotated_position[sym][p] = b
                        break
                else:
                    assert False, 'not a permutation'
        return rotated_position
        
    def get_data(self, sizes, vectordata):
        polys = self.create_cells(sizes)
        
        data_cells_visible_faces = [[f.id for f in cell.faces if f.type == 'face'] for cell in polys.cells]
        polys_verts = [v.point for v in polys.verts]
        data_texranges_mosaic = [list(get_texcoords_range(polys_verts, self.normals[sym])) for sym in self.faces]
        data_rotated_position = self.get_rotated_position(polys.cells)
        data_cell_indices = [cell.indices for cell in polys.cells]
        data_cell_centers = [vectordata.toindex(cell.center()) for cell in polys.cells]
        data_pick_polygons = list(self.gl_pick_polygons(polys.cells, vectordata))
        blocksdata = self.gl_block_data(polys, vectordata)
        
        return {
                'rotated_position': data_rotated_position,
                'cell_indices': data_cell_indices,
                'cell_centers': data_cell_centers,
                'cells_visible_faces': data_cells_visible_faces,
                'texranges_mosaic': data_texranges_mosaic,
                'blocksdata': blocksdata,
                'facesdata': data_pick_polygons,
            }
        
    def gl_block_data(self, polys, vectordata):
        polys_faces = polys.faces
        def gl_block_data_cell(cell):
            def is_visible(vert):
                for hv in vert.halfverts:
                    he = hv.halfedge
                    hf = he.halfface
                    if hv is he.halfverts[1] and hf.hdim is cell and not hf.face.type.endswith('_removed'):
                        return True
                return False
            def gen_cedges(cell_halffaces):
                edgesdone = []
                for hf in cell_halffaces:
                    for he1 in hf.halfedges:
                        if he1.edge not in edgesdone:
                            he3 = he1.other
                            hf3_face = he3.halfface.face
                            edgesdone.append(he1.edge)
                            if not (hf.face.type.endswith('_removed') and hf3_face.type.endswith('_removed')):
                                v1_point = vectordata.toindex(he1.verts[1].point)
                                v3_point = vectordata.toindex(he3.verts[1].point)
                                yield polys_faces.index(hf.face), polys_faces.index(hf3_face), v1_point, v3_point
            def gen_vhalfface(halfface, vert):
                for hv in vert.halfverts:
                    he = hv.halfedge
                    hf = he.halfface
                    if hv is he.halfverts[1] and hf is halfface:
                        break
                else:
                    assert False
                while True:
                    assert hv.vert is vert
                    yield polys_faces.index(hf.face)
                    he = hv.halfedge.other
                    hv = he.halfverts[0].other
                    assert hv.vert is vert
                    hf = he.halfface
                    if hf is halfface:
                        break
            def gen_cverts(cell_halffaces):
                vertsdone = []
                for hf in cell_halffaces:
                    for he1 in hf.halfedges:
                        he3 = he1.other
                        hf3 = he3.halfface
                        v1 = he1.verts[1]
                        v3 = he3.verts[1]
                        if v1 not in vertsdone:
                            vertsdone.append(v1)
                            if is_visible(v1):
                                vertdata = [f for f in gen_vhalfface(hf, v1)]
                                yield vectordata.toindex(v1.point), vertdata
                        if v3 not in vertsdone:
                            vertsdone.append(v3)
                            if is_visible(v3):
                                vertdata = [f for f in gen_vhalfface(hf3, v3)]
                                yield vectordata.toindex(v3.point), vertdata
            def gen_cfaces(cell_halffaces):
                c_faces_label = []
                c_faces_black = []
                for hf in cell_halffaces:
                    verts = [vectordata.toindex(v.point) for v in hf.verts]
                    if hf.face.type == 'face':
                        c_faces_label.append((polys_faces.index(hf.face), self.faces.index(hf.face.id), verts))
                    elif hf.face.type == 'cut':
                        c_faces_black.append((polys_faces.index(hf.face), verts))
                    else:
                        assert hf.face.type.endswith('_removed')
                return c_faces_label, c_faces_black
            c_faces_label, c_faces_black = list(gen_cfaces(cell.halffaces))
            c_edges = list(gen_cedges(cell.halffaces))
            c_verts = list(gen_cverts(cell.halffaces))
            return c_faces_label, c_faces_black, c_edges, c_verts
        return [gl_block_data_cell(cell) for cell in polys.cells]
        
    def gl_pick_polygons(self, cells, vectordata):
        black_faces = []
        for cellidx, cell in enumerate(cells):
            cell_facesdata = []
            for halfface in cell.halffaces:
                if halfface.face.type == 'face':
                    symbol = halfface.face.id
                    face = self.faces.index(symbol)
                    visible_face_indices = [i for i, he in enumerate(halfface.halfedges)
                                          if he.other.halfface.face.type == 'face']
                else:
                    if halfface.face in black_faces:
                        continue
                    black_faces.append(halfface.face)
                    face = halfface.face.id
                edges_iverts = [vectordata.toindex(he.verts[0].point) for he in halfface.halfedges]
                if halfface.face.type != 'face':
                    cell_facesdata.append([face, None, edges_iverts])
                    continue
                if len(visible_face_indices) == 0:
                    picktype = 0
                elif len(visible_face_indices) == 1:
                    if len(edges_iverts) == 3:
                        picktype = 1
                    elif len(edges_iverts) == 4:
                        picktype = 2
                    elif len(edges_iverts) == 6:
                        picktype = 4
                    else:
                        assert False, len(edges_iverts)
                    i = visible_face_indices[0]
                    edges_iverts = edges_iverts[i:] + edges_iverts[:i]
                elif len(visible_face_indices) == 2:
                    assert 3 <= len(edges_iverts) <= 5, len(edges_iverts)
                    i = visible_face_indices[0]
                    if len(edges_iverts) < 5:
                        picktype = 3
                    else:
                        picktype = 5
                    if i+1 == visible_face_indices[1]:
                        edges_iverts = edges_iverts[i:] + edges_iverts[:i]
                    elif visible_face_indices == [0, len(edges_iverts)-1]:
                        i = visible_face_indices[1]
                        edges_iverts = edges_iverts[i:] + edges_iverts[:i]
                    else:
                        picktype = -1
                else:
                    picktype = -1
                cell_facesdata.append([face, picktype, edges_iverts])
            yield cell_facesdata
                

class Dedup:
    def __init__(self):
        self.dedup_data = defaultdict(list)
        self.cnt_dups = 0
        self.cnt_values = 0
        
    def _float(self, value):
        self.cnt_values += 1
        dddlt = self.dedup_data['f']
        value = roundeps(value)
        if value == 0.0:  # 0.0 == -0.0
            value = 0.0  # for reproducible build
        didx = bisect_left(dddlt, value)
        if didx < len(dddlt) and dddlt[didx] == value:
            self.cnt_dups += 1
            return dddlt[didx], True, 'f'
        else:
            dddlt.insert(didx, value)
            return value, False, 'f'
            
    def _recursion(self, value, iterable):
        dedup = True
        vtypes = []
        for k, v in iterable:
            v, d, t = self.dedup(v)
            dedup = dedup and d
            vtypes.append(t)
            value[k] = v
        return dedup, '('+''.join(vtypes)+')'
        
    def _replace(self, value, dedup, vtype):
        self.cnt_values += 1
        dddlt = self.dedup_data[vtype]
        try:
            return self._replace_bisect(dddlt, value, dedup, vtype)
        except TypeError:
            return self._replace_index(dddlt, value, dedup, vtype)
            
    def _replace_index(self, dddlt, value, dedup, vtype):
        if not dedup:
            #assert value not in dddlt
            dddlt.append(value)
            return value, False, vtype
        try:
            didx = dddlt.index(value)
        except ValueError:
            dddlt.append(value)
            return value, False, vtype
        else:
            dvalue = dddlt[didx]
            self.cnt_dups += 1
            return dvalue, True, vtype
        
    def _replace_bisect(self, dddlt, value, unused_dedup, vtype):
        didx = bisect_left(dddlt, value)
        try:
            dvalue = dddlt[didx]
        except IndexError:
            dddlt.append(value)
            return value, False, vtype
        if value == dvalue:
            self.cnt_dups += 1
            return dvalue, True, vtype
        else:
            dddlt.insert(didx, value)
            return value, False, vtype
            
    def dedup(self, value):
        if type(value) is dict:
            value = {(sys.intern(k) if type(k) is str else k):v for k,v in value.items()}
            dedup, vtype = self._recursion(value, value.items())
            return self._replace(value, dedup, 'd' + vtype)
        elif type(value) is list:
            dedup, vtype = self._recursion(value, enumerate(value))
            return self._replace(tuple(value), dedup, 'a' + vtype)
        elif type(value) is tuple:
            value = list(value)
            dedup, vtype = self._recursion(value, enumerate(value))
            return self._replace(tuple(value), dedup, 'a' + vtype)
        elif type(value) is float:
            return self._float(value)
        elif type(value) is int:
            return value, True, 'i'
        elif type(value) is str:
            return sys.intern(value), True, 's'
        elif type(value) is bool:
            return value, True, 'b'
        elif value is None:
            return value, True, 'n'
        else:
            assert False, type(value)
            
def pool_functions(parallel, reproducible):
    pool = None
    if reproducible:
        print('warning: using undocumented interface in module pickle for reproducible build')
    try:
        if parallel > 1:
            pool = Pool(processes=parallel, initializer=make_dumps, initargs=[reproducible])
    except OSError as e:
        print('process pool not available ({}):'.format(e))
        print('  deactivating multiprocessing')
    sys.stdout.flush() # when Pool(…) fails this line is 
    if pool is not None:
        return pool.imap_unordered
    else:
        make_dumps(reproducible)
        return map
        
#XXX: speedup parallel builds, longer jobs first
fileorder = 'b111 d000000 b101 b000 b010 p111111 p101111 b100 t1111 b110 p010000 b011 p000000 t0000 b001 t1011 t0100'.split()

def get_datafilename(Factory, sizes):
    if DEBUG_MODELFAST:
        return 'f{:02}{}{}'.format(modeldefs.index(Factory)+1, Factory.fileclass, ''.join(str(s-1) for s in sizes))
    else:
        sizes = Factory.fileclass + ''.join(str((s-1)%2) for s in sizes)
        try:
            priority = fileorder.index(sizes) + 1
        except ValueError:
            priority = 99
        return 'd{:02}'.format(priority) + sizes
        
def enum_modelfiles(Factory):
    def tuples(maxlen, part=()):
        if maxlen == len(part):
            yield part
        else:
            for i in range(minsize, maxsize+1):
                yield from tuples(maxlen, part+(i,))
    for size in tuples(len(Factory.sizenames)):
        norm_size, sizes = Factory.norm_sizes(size)
        if sizes is None:
            continue
        filename = get_datafilename(Factory, sizes)
        yield filename, sizes, size, norm_size
            
def pool_enum_modelfiles(dirname, testfunc):
    ignored = []
    modelfiles = []
    for Factory in modeldefs:
        for filename, *unused in enum_modelfiles(Factory):
            filename = os.path.join(dirname, filename)
            if filename in modelfiles or filename in ignored:
                continue
            if testfunc(filename):
                modelfiles.append(filename)
            else:
                ignored.append(filename)
    if modelfiles:
        for filename in sorted(ignored):
            print('skipping', filename)
    return sorted(modelfiles)
    
def pool_create_modelfiledata(path):
    filename = os.path.basename(path)
    savedata = {}
    vectordata = VectorData()
    for Factory in modeldefs:
        sizes_list = []
        factory = None
        for _filename, sizes, *unused in enum_modelfiles(Factory):
            if _filename != filename:
                continue
            if sizes in sizes_list:
                continue
            sizes_list.append(sizes)
            if factory is None:
                factory = ModelFactory(Factory)
            savedata.setdefault(factory.type, {})[sizes] = factory.get_data(sizes, vectordata)
    return savedata, vectordata.vectors, vectordata.cnts, vectordata.dups
    
def format_seconds(seconds):
    if seconds > 60:
        return '{:.0f}m {:.2f}s'.format(*divmod(seconds, 60))
    else:
        return '{:.2f}s'.format(seconds)
    
def pool_check_compare_modelfiledata(check_savedata, check_vectors, savedata, vectors):
    if savedata == check_savedata and vectors == check_vectors:
        return 'check: pass full'
        
    for mtype, check_value_mtype in list(check_savedata.items()):
        if mtype not in savedata:
            del check_savedata[mtype]
        else:
            value_mtype = savedata[mtype]
            for sizes, check_value_sizes in list(check_value_mtype.items()):
                if sizes not in value_mtype:
                    del check_value_mtype[sizes]
                else:
                    value_sizes = value_mtype[sizes]
                    for attr in 'block_polygons', 'pick_polygons':  # older versions may have this attrs
                        with suppress(KeyError):
                            del check_value_sizes[attr]
                    for attr in 'facesdata', 'blocksdata', 'normals':
                        with suppress(KeyError):
                            del check_value_sizes[attr]
                        del value_sizes[attr]
    for mtype, value_mtype in list(savedata.items()):
        if mtype not in check_savedata:
            del savedata[mtype]
        else:
            for sizes in list(value_mtype.keys()):
                if sizes not in check_savedata[mtype]:
                    del value_mtype[sizes]
    if savedata == check_savedata:
        return 'check: pass weak'
    return 'check: data different'
    
def write_diff(filename, check_savedata, savedata):
    from difflib import unified_diff
    from pprint import pformat
    columns = int(os.environ.get('COLUMNS') or '80') - 1
    diff = unified_diff(pformat(check_savedata, width=columns).splitlines(),
                        pformat(savedata, width=columns).splitlines(),
                        fromfile=filename, tofile='<generated by {}>'.format(__file__),
                        n=3, lineterm='')
    filename += '.diff'
    isempty = True
    with open(filename, 'wt') as file:
        for line in diff:
            print(line, file=file)
            isempty = False
    if isempty and os.path.exists(filename):
        os.remove(filename)
            
def pool_create_modelfile(filename, pickle_protocol, check):
    import time
    seconds = time.process_time()
    
    if check:
        import pickle
        import pybiklib.model
        try:
            with open(filename, 'rb') as datafile:
                check_savedata = pickle.load(datafile)
                check_vectors = pybiklib.model.Model.read_vectors(datafile)
        except FileNotFoundError:
            check_savedata = {}
            check_vectors = []
        check = (check_savedata, check_vectors)
    else:
        check = None
        
    savedata, vectors, cnt_vectors, dup_vectors = pool_create_modelfiledata(filename)
    vals = dups = ''
    if not DEBUG_MODELFAST:
        dedup = Dedup()
        savedata = dedup.dedup(savedata)[0]
        if DEBUG_MSG:
            vals = '\n  vals: %6s  vec: %6s' % (dedup.cnt_values, cnt_vectors)
            dups = '\n  dups: %6s       %6s' % (dedup.cnt_dups, dup_vectors)
    
    bsavedata = dumps(savedata, pickle_protocol)
    datasize = ['{:.1f} kb'.format(len(bsavedata) / 1000), '---']
    vectors = array('f', chain.from_iterable(vectors))
    vectorssize = '{:.1f} kb'.format(len(vectors) * vectors.itemsize / 1000)
    if not DEBUG_MODELFAST:
        bsavedata = pickletools.optimize(bsavedata)
        datasize[1] = '{:.1f} kb'.format(len(bsavedata) / 1000)
    seconds = time.process_time() - seconds
    
    if check is None:
        vlen = array('I', [len(vectors)])
        if sys.byteorder != filebyteorder:
            vlen.byteswap()
            vectors.byteswap()
        with open(filename, 'wb') as datafile:
            datafile.write(bsavedata)
            vlen.tofile(datafile)
            vectors.tofile(datafile)
        message = 'generated'
    else:
        message = pool_check_compare_modelfiledata(check_savedata, check_vectors, savedata, vectors)
        write_diff(filename, check_savedata, savedata)
    return filename, seconds, '{} {:{}} ({:>9}, {:>9}), {:>9}, {:7} vectors {:>8}{}{}'.format(
                    message, filename, len(os.path.dirname(filename))+11,
                    datasize[0], datasize[1], format_seconds(seconds), len(vectors) / 3, vectorssize,
                    vals, dups)
    
def pool_create_indexdata():
    savedata_type = {}
    savedata_types = []
    savedata_size = {}
    savedata_sizes = {}
    savedata_facenames = []
    savedata = {'type': savedata_type, 'types': savedata_types,
                'normsize': savedata_size, 'sizes': savedata_sizes,
                'facenames': savedata_facenames,
               }
    facekeys = []
    for Factory in modeldefs:
        factory = ModelFactory(Factory)
        savedata_type[factory.type] = factory.getattrs()
        savedata_types.append(factory.type)
        savedata_size_type = {}
        savedata_sizes_type = {}
        savedata_size[factory.type] = savedata_size_type
        savedata_sizes[factory.type] = savedata_sizes_type
        for facekey, facename in zip(factory.facekeys, factory.facenames):
            if facekey not in facekeys:
                facekeys.append(facekey)
                savedata_facenames.append((facekey, facename))
        for filename, sizes, size, norm_size in enum_modelfiles(Factory):
            savedata_size_type[size] = norm_size
            if norm_size in savedata_sizes_type:
                assert [sizes, filename] == savedata_sizes_type[norm_size]
            else:
                savedata_sizes_type[norm_size] = [sizes, filename]
        size_range = ((min(vals), max(vals)) for vals in zip(*savedata_size_type.keys()))
        defaultsize = savedata_type[factory.type]['defaultsize']
        defaultsize = tuple(max(smin, min(s, smax)) for s, (smin, smax) in zip(defaultsize, size_range))
        savedata_type[factory.type]['defaultsize'] = defaultsize
    return savedata
    
def pool_check_compare_indexdata(check, savedata):
    if check == savedata:
        return 'check: pass full'
        
    def remove_mtype(data, mtype):
        del data['normsize'][mtype]
        del data['sizes'][mtype]
        del data['type'][mtype]
    for mtype in check['types']:
        if mtype not in savedata['types']:
            remove_mtype(check, mtype)
    for mtype in savedata['types']:
        if mtype not in check['types']:
            remove_mtype(savedata, mtype)
    for attr in 'normsize','sizes',:
        for mtype, check_value in check[attr].items():
            savedata_value = savedata[attr][mtype]
            for key in list(check_value.keys()):
                if key not in savedata_value:
                    del check_value[key]
            
    if check == savedata:
        return 'check: pass weak'
    return 'check: data different'
        
def pool_create_indexfile(filename, pickle_protocol, check):
    import time
    seconds = time.process_time()
    
    if check:
        import pybiklib.model
        pybiklib.model.Model.load_index()
        check = pybiklib.model.Model.cache_index
    else:
        check = None
    savedata = pool_create_indexdata()
    vals = dups = ''
    if not DEBUG_MODELFAST:
        dedup = Dedup()
        savedata = dedup.dedup(savedata)[0]
        if DEBUG_MSG:
            vals = '\n  vals: %6s' % dedup.cnt_values
            dups = '\n  dups: %6s' % dedup.cnt_dups
    bsavedata = dumps(savedata, pickle_protocol)
    datasize = ['{:.1f} kb'.format(len(bsavedata) / 1000), '---']
    if not DEBUG_MODELFAST:
        bsavedata = pickletools.optimize(bsavedata)
        datasize[1] = '{:.1f} kb'.format(len(bsavedata) / 1000)
    seconds = time.process_time() - seconds
    
    if check is None:
        with open(filename, 'wb') as datafile:
            datafile.write(bsavedata)
        message = 'generated'
    else:
        message = pool_check_compare_indexdata(check, savedata)
        write_diff(filename, check, savedata)
    return filename, seconds, '{} {:{}} ({:>9}, {:>9}), {:>9}{}{}'.format(
                    message, filename, len(os.path.dirname(filename))+11,
                    datasize[0], datasize[1], format_seconds(seconds),
                    vals, dups)
    
def pool_run(args):
    func, *args = args
    return func(*args)
    
def get_indexfilename(dirname):
    return os.path.join(dirname, 'f00index' if DEBUG_MODELFAST else 'd00index')
    
def create_modeldata(dirname, testfunc=None, parallel=1, pickle_protocol=-1,
                              reproducible=False, check=False):
    # prepare jobs
    if testfunc is None:
        testfunc = lambda arg: True
    if DEBUG_INDEXONLY:
        modelfiles = []
    else:
        modelfiles = pool_enum_modelfiles(dirname, testfunc)
    jobs = [(pool_create_modelfile, m, pickle_protocol, check) for m in modelfiles]
    indexfilename = get_indexfilename(dirname)
    jobs.append((pool_create_indexfile, indexfilename, pickle_protocol, check))
    # run jobs
    if parallel is True:
        parallel = cpu_count()
    if not parallel or parallel < 1:
        parallel = 1
    realparallel = min(parallel, len(jobs))
    print('using {} / {} processes'.format(realparallel, parallel))
    imap_model = pool_functions(realparallel, reproducible)
    result = []
    for filename, seconds, lines in imap_model(pool_run, jobs):
        result.append((filename, seconds))
        print(lines)
        sys.stdout.flush()
    if DEBUG_MSG and not DEBUG_MODELFAST:
        result.sort(key=lambda fs: fs[1], reverse=True)
        if result != sorted(result, key=lambda fs: fs[0]):
            print('warning: reorder jobs to optimize parallel build')
            for filename, seconds in result:
                print(' ', os.path.basename(filename), format_seconds(seconds))
    
    
if __name__ == '__main__':
    minsize = 1
    maxsize = 5
    create_modeldata('data/models',
                     check=True,
                    )
    




# -*- coding: utf-8 -*-
# cython: profile=False

#  Copyright © 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=C0326,W0614
# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__
#px/__compiled = True
__compiled = False

import sys

from pybiklib import config as config_

#px+from qt cimport *
#px/cimport qt as Qt
from .debug_purepython import *

#px/from libc.stdio cimport printf
def printf(fmt, *args): print(fmt % args, end='')
#px/from libc.stdio cimport puts
puts = print

#px/cdef enum: #
if True:
    DEBUG_MSG = 0x0002
    DEBUG_MSGEXT = 0x0020
#px+cdef long debug
debug = 0
    
def set_debug_flags(module):
    global debug
    if module.DEBUG_MSG:   debug |= DEBUG_MSG
    if module.DEBUG_MSGEXT: debug |= DEBUG_MSGEXT
    

#### Helpers (same as in module qt) ####

#pxm-FUNC P
def q2str(qstr:str)->'':  return qstr
    #px+cdef QByteArray data = qstr.toUtf8()
    #px+return data.data()[:data.size()].decode('utf-8')
#pxm-FUNC P
def str2q(pstr)->str:  return pstr
    #px+data = pstr.encode('utf-8')
    #px+return fromUtf8(<char*>data, len(data))
    

#### . ####

def get_gl_variant():
    #px+cdef int i
    # This function requires that the app instance is already created
    i = openGLModuleType()
    if i == LibGL:
        return 'ogl'
    elif i == LibGLES:
        return 'es2'
    else:
        print('Unknown openGLModuleType', i)
        return ''
        
def exec_application(main):
    #px-
    global app #http://pyqt.sourceforge.net/Docs/PyQt5/gotchas.html#crashes-on-exit
    
    #setAttribute(Qt.AA_X11InitThreads)
    #setAttribute(Qt.AA_UseDesktopOpenGL)
    #setAttribute(Qt.AA_UseSoftwareOpenGL)
    
    #px+try:
    #px+    byteslist = [arg.encode(sys.getdefaultencoding()) for arg in main.opts.ui_args]
    #px+except Exception:
    #px+    sys.excepthook(*sys.exc_info())
    #px+    byteslist = []
    #px+args_buffer = b'\0'.join(byteslist) + b'\0'
    #px+
    #px+ptr = int(<size_t><void*><char*>args_buffer)
    #px+ptrlist = []
    #px+for arg in byteslist:
    #px+    ptrlist.append(ptr)
    #px+    ptr += len(arg)+1
    #px+ptrlist.append(0)
    #px+
    #px+import struct
    #px+argv_buffer = struct.pack('{}P'.format(len(ptrlist)), *ptrlist)
    #px+cdef int argc = len(byteslist)
    #px+cdef char **argv = <char**><void*><char*>argv_buffer
    #px/cdef QApplication *app = new QApplication(argc, argv)
    app = QApplication(main.opts.ui_args)
    
    qargs = app.arguments()
    #px/if qargs.size() > 1:
    if len(qargs) > 1:
        print('Unknown arguments:')
        #px/for i in range(1, qargs.size()):
        for i in range(1, len(qargs)):
            print(' ', q2str(qargs[i]))
        return 1
    
    app.setOrganizationName(str2q(config_.PACKAGE))
    app.setApplicationName(str2q(config_.APPNAME))
    app.setApplicationVersion(str2q(config_.VERSION))
    app.setWindowIcon(QIcon(str2q(config_.APPICON_FILE)))
    
    # initialize translation
    language = q2str(QLocale_system().name())
    # standard Qt translation, used for e.g. standard buttons and shortcut names
    # keep the translator object for the lifetime of the app object, so that the translation works
    #px/cdef QTranslator translator
    translator = QTranslator()
    translator.load(str2q("qt_" + language), location(TranslationsPath))
    #px/app.installTranslator(&translator)
    app.installTranslator(translator)
    main.cb_app_post_create(language)
    
    #XXX: needed for threaded rendering, prevents blocking on quit
    #px/connect(app, &aboutToQuit, app, &sync, Qt.DirectConnection)
    app.aboutToQuit.connect(app.sync, type=Qt.DirectConnection)
    
    while True:
        pyapp = main.cb_run_app()
        if pyapp is None:
            break
        pyapp.cb_before_mainloop()
        app.exec()
        pyapp.cb_after_mainloop()
    del app
    
#px+depends = [[DEPENDS]]
#px+aliases = [[ALIASES]]





#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
from ast import literal_eval
from io import StringIO
from collections import OrderedDict
from contextlib import suppress

from . import config


current_settings_version = 3

class KeyStore:
    def __init__(self):
        self.schema = None
        self.filename = None
        self.keystore = {}
        self.callback = lambda unused_key: None
        self.globtree = GlobTree()
        self.extra_lines = []
        
    def clone_key(self, keyfrom, keyto):
        self.schema[keyto] = self.schema[keyfrom]
        if keyfrom in self.keystore and keyto not in self.keystore:
            self.keystore[keyto] = self.keystore[keyfrom]
            
    def get_default(self, key):
        return self.schema[key][0]
    def get_validator(self, key):
        return self.schema[key][1]
        
    def get_range(self, key):
        validator = self.get_validator(key)
        if not isinstance(validator, (tuple, list)):
            raise ValueError('{} is not a range'.format(key))
        return validator
    
    def get_value(self, key):
        try:
            return self.keystore[key]
        except KeyError:
            return self.get_default(key)
            
    def get_nick(self, key):
        value = self.get_value(key)
        validator = self.get_validator(key)
        if not isinstance(validator, list):
            raise ValueError('{} is not an enum'.format(key))
        return validator[value]
        
    def set_value(self, key, value):
        if key not in self.keystore:
            if value != self.get_default(key):
                self.keystore[key] = value
                self.callback(key)
        elif self.keystore[key] != value:
            self.keystore[key] = value
            self.callback(key)
        
    def set_nick(self, key, nick):
        validator = self.get_validator(key)
        if not isinstance(validator, list):
            raise ValueError('{} is not an enum'.format(key))
        value = validator.index(nick)
        return self.set_value(key, value)
        
    def del_value(self, key):
        try:
            value = self.keystore[key]
        except KeyError:
            pass # already the default value
        else:
            del self.keystore[key]
            if value != self.get_default(key):
                self.callback(key)
            else:
                self.callback(None)
        
    def find_match(self, key):
        def match(key, pattern):
            key = key.split('.')
            pattern = pattern.split('.')
            for k, p in zip(key, pattern):
                if k != p != '*':
                    return False
            return True
        for pattern in self.schema:
            if match(key, pattern):
                return pattern
        return None
        
    def read_settings(self, filename):
        from .schema import schema, deprecated
        self.schema = schema
        self.filename = filename
        
        # read settings
        lines = []
        if self.filename:
            dirname = os.path.dirname(self.filename)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)
            try:
                with open(self.filename, 'rt', encoding='utf-8') as settings_file:
                    text = settings_file.read()
                    lines = text.splitlines()  # line breaks are not included
            except FileNotFoundError:
                pass
            except Exception:
                sys.excepthook(*sys.exc_info())
                
        for line in lines:
            try:
                key, strvalue = line.split('=', 1)
            except ValueError:
                # discard invalid lines
                continue
                
            key = key.strip()
            pattern = self.find_match(key)
            if pattern is None:
                # keep unknown keys, they may be settings of a different version
                self.extra_lines.append(line)
                continue
            self.schema[key] = self.schema[pattern]
            
            strvalue = strvalue.strip()
            try:
                value = literal_eval(strvalue)
            except (ValueError, SyntaxError):
                # discard invalid values
                continue
                
            # translate enums and validate values, discard invalid values
            validator = self.get_validator(key)
            if validator is deprecated:
                pass
            elif isinstance(validator, list):
                try:
                    value = validator.index(value)
                except ValueError:
                    continue
            elif isinstance(validator, tuple):
                if not validator[0] <= value <= validator[1]:
                    continue
            elif validator is not None:
                if not validator(value):
                    continue
                    
            self.keystore[key] = value
            
        self.build_globtree()
        
    def build_globtree(self):
        keys = list(self.schema.keys())
        for key in keys:
            subkeys = key.split('.')
            if '*' in subkeys:
                self.globtree.insert_key(subkeys)
                
    def resolve_glob(self, key):
        subkeys = key.split('.')
        clone = self.globtree.find_key(subkeys)
        assert clone is not None, subkeys
        globkey = '.'.join(clone)
        assert globkey != key
        self.clone_key(globkey, key)
        
    def dump(self, file, all=False):    # pylint: disable=W0622
        keydict = self.schema if all else self.keystore
        for key in sorted(keydict.keys()):
            if '*' in key:
                continue
            try:
                # translate enums
                value = self.get_nick(key)
            except ValueError:
                value = self.get_value(key)
            print(key, '=', repr(value), file=file)
        if self.extra_lines:
            #XXX: this line is discarded when loading the settings file
            #     because it does not contain a '='
            print('# comments and keys unknown by version {}:'.format(config.VERSION), file=file)
            for line in self.extra_lines:
                print(line, file=file)
        
    def write_settings(self):
        if not self.filename:
            return
        buf = StringIO()
        self.dump(buf)
        tmpfilename = self.filename + '.tmp'
        with suppress(OSError):
            os.remove(tmpfilename)
        try:
            with open(tmpfilename, 'wt', encoding='utf-8') as settings_file:
                settings_file.write(buf.getvalue())
            os.replace(tmpfilename, self.filename)
        except Exception:
            sys.excepthook(*sys.exc_info())
            
    def get_by_suffix(self, key):
        if key.endswith('_nick'):
            key = key[:-5]
            func = self.get_nick
        elif key.endswith('_range'):
            key = key[:-6]
            func = self.get_range
        elif key.endswith('_default'):
            key = key[:-8]
            func = self.get_default
        else:
            func = self.get_value
        try:
            return func(key)
        except KeyError:
            self.resolve_glob(key)
            return func(key)
            
    def set_by_suffix(self, key, value):
        if key.endswith('_nick'):
            key = key[:-5]
            func = self.set_nick
        else:
            func = self.set_value
        try:
            func(key, value)
        except KeyError:
            self.resolve_glob(key)
            func(key, value)
            
    def del_by_suffix(self, key):
        try:
            self.del_value(key)
        except KeyError:
            self.resolve_glob(key)
            self.del_value(key)
            
        
class GlobTree:
    __slots__ = 'nodes', 'leaves'
    
    def __init__(self):
        self.nodes = OrderedDict()
        self.leaves = []
        
    def insert_key(self, subkeys):
        subkey, *subkeys = subkeys
        if subkeys:
            try:
                node = self.nodes[subkey]
            except KeyError:
                node = GlobTree()
                self.nodes[subkey] = node
            node.insert_key(subkeys)
        elif subkey not in self.leaves:
            self.leaves.append(subkey)
        else:
            assert False
        
    def find_key(self, subkeys):
        subkey, *subkeys = subkeys
        if subkeys:
            for subkey in subkey, '*':
                try:
                    node = self.nodes[subkey]
                except KeyError:
                    continue
                clone = node.find_key(subkeys)
                if clone is not None:
                    return [subkey] + clone
            return None
        elif subkey in self.leaves:
            return [subkey]
        elif '*' in self.leaves:
            return ['*']
        else:
            return None
            
        
class Settings:
    __slots__ = ()
    keystore = KeyStore()
    dump = keystore.dump
    
    @classmethod
    def reset(self):
        self.keystore.keystore.clear()
        
    def load(self, filename):
        self.keystore.read_settings(filename)
        
        version = self['version']
        if version != current_settings_version:
            self['version'] = current_settings_version
        return version
            
    @classmethod
    def connnect(self, callback):
        self.keystore.callback = callback
        
    @classmethod
    def disconnect(self):
        self.keystore.callback = lambda unused_key: None
        
    @classmethod
    def flush(self):
        self.keystore.write_settings()
        
    @staticmethod
    def norm_key(key):
        if type(key) is int:
            return str(key)
        elif type(key) is tuple:
            def cvt_dot(keys):
                for key in keys:
                    if type(key) is int:
                        yield str(key)
                    elif type(key) is tuple:
                        yield '_'.join(str(v) for v in key)
                    else:
                        yield from key.split('.')
            return '.'.join(cvt_dot(key))
        return key
        
    def __getitem__(self, key):
        key = self.norm_key(key)
        return self.keystore.get_by_suffix(key)
        
    def __setitem__(self, key, value):
        key = self.norm_key(key)
        self.keystore.set_by_suffix(key, value)
        
    def __delitem__(self, key):
        key = self.norm_key(key)
        self.keystore.del_by_suffix(key)
        
    

settings = Settings()




# -*- coding: utf-8 -*-

#  Copyright © 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


from qt cimport *


ctypedef void CallbackErrorList(const QmlErrorList &) nogil


#### QtCore ####

#QList[QQmlError]
cdef extern from "<QtCore/QList>":
    cdef cppclass QmlErrorList "QList<QQmlError>":
        int size() const
        const QQmlError& at(int i) const
        
#QObject
cdef extern from "<QtCore/QObject>" namespace "QObject":
    Connection connect(const QQmlEngine *sender, CallbackErrorList *signal, CallbackErrorList slot) nogil
    

#### QtQml ####

#QQmlComponent
cdef extern from "<QtQml/QQmlComponent>":
    cdef cppclass QQmlComponent:
        QQmlComponent(QQmlEngine *, QObject *parent)
        void loadUrl(const QUrl &url)
        void loadUrl(const QUrl &url, CompilationMode mode)
        cbool isError() const
        cbool isLoading() const
        QmlErrorList errors() const
        QObject *beginCreate(QQmlContext *)
        void completeCreate()
cdef extern from "<QtQml/QQmlComponent>" namespace "QQmlComponent":
    cdef enum CompilationMode:
        PreferSynchronous, Asynchronous
        
#QQmlContext
cdef extern from "<QtQml/QQmlContext>":
    cdef cppclass QQmlContext:
        void setContextProperty(const QString &name, QObject *value)
        
#QQmlEngine
cdef extern from "<QtQml/QQmlEngine>":
    cdef cppclass QQmlEngine:
        QQmlEngine()
        void setOutputWarningsToStandardError(cbool enabled)
        QQmlContext *rootContext() const
        void clearComponentCache()
cdef extern from "<QtQml/QQmlEngine>" namespace "QQmlEngine":
    void warnings(const QmlErrorList &warnings) nogil
        
#QQmlError
cdef extern from "<QtQml/QQmlError>":
    cdef cppclass QQmlError:
        QString description() const
        QUrl url() const
        int line() const
        int column() const
        
        
#### QtQuick ####

#QQuickItem
cdef extern from "<QtQuick/QQuickItem>":
    cdef cppclass QQuickItem(QObject):
        qreal x() nogil const
        qreal y() nogil const
        qreal width() nogil const
        qreal height() nogil const
        QQuickItem *parentItem() nogil const
        QQuickWindow *window() nogil const
        
#QWindow
cdef extern from "<QtGui/QWindow>":
    cdef cppclass QWindow:
        QSurfaceFormat format() nogil const
        void resize(int w, int h) nogil
        cbool close() nogil
        int width() nogil const
        int height() nogil const
        cbool setMouseGrabEnabled(cbool grab) nogil
        #void setParent(QWindow *parent)
        void setTransientParent(QWindow * parent) nogil
#QQuickWindow
cdef extern from "<QtQuick/QQuickWindow>":
    cdef cppclass QQuickWindow(QWindow):
        void update() nogil
        void setCursor(const QCursor &cursor) nogil
        void unsetCursor() nogil
        void setPersistentOpenGLContext(cbool persistent) nogil
        void setClearBeforeRendering(cbool enabled) nogil
        QImage grabWindow()
        int x() nogil const
        int y() nogil const
        QQuickItem *contentItem() nogil const
        void setRenderTarget(QOpenGLFramebufferObject *fbo) nogil
#QQuickView
cdef extern from "<QtQuick/QQuickView>":
    cdef cppclass QQuickView(QQuickWindow):
        QQuickView(QQmlEngine *engine, QQuickWindow *parent)
        void setSource(const QUrl &url)
        void setTitle(const QString &title)
        QmlErrorList errors() const
        void setResizeMode(ResizeMode)
        void show()
cdef extern from "<QtQuick/QQuickView>" namespace "QQuickView":
    cdef enum ResizeMode:
        SizeViewToRootObject, SizeRootObjectToView
    void beforeSynchronizing() nogil
    void beforeRendering() nogil
        
        





#-*- coding:utf-8 -*-

#  Copyright © 2009-2015, 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
from collections import namedtuple

from .debug import debug


class MoveData (namedtuple('_MoveData', 'axis slice dir')):     # pylint: disable=W0232
    # axis  = 0...
    # slice = 0...dim-1 or -1 for all slices
    # dir   = 0 or 1
    __slots__ = ()
    def inverted(self):
        return self._replace(dir=not self.dir)  # pylint: disable=E1101
        

def normalized_dirs(model, axis, dirs):
    symmetry = model.symmetries[axis]
    dirs = dirs % symmetry
    if dirs > symmetry // 2:
        dirs -= symmetry
    return dirs
    
class MoveDataPacked (namedtuple('_MoveDataPacked', 'axis slice dirs')):
    # axis  = 0...
    # slice = 0...dim-1, (-1 for all slices not allowed in some functions)
    # dirs = -symmetry/2...+symmetry/2, >0 CW <0 CCW
    __slots__ = ()
    
    def __new__(self, maxis, mslice, mdirs, model=None):
        if model is not None:
            mdirs = normalized_dirs(model, maxis, mdirs)
        return super().__new__(self, maxis, mslice, mdirs)
        
    def length(self):
        return abs(self.dirs)
        
    def add_dirs(self, dirs, model):
        return self._replace(dirs=normalized_dirs(model, self.axis, self.dirs+dirs))
        
    def unpack(self):
        mdir = self.dirs < 0
        for unused in range(self.length()):
            yield self.axis, self.slice, mdir
            
        
class _MoveQueueBase:
    class MoveQueueItem:
        def __init__(self, data, *, mark_after=False, code=None):
            assert isinstance(data, (MoveData, MoveDataPacked))
            self.data = data
            self.mark_after = mark_after
            self.code = code
        axis = property(lambda self: self.data.axis)
        slice = property(lambda self: self.data.slice)
        dir = property(lambda self: self.data.dir)
        def copy(self):
            return self.__class__(self.data, mark_after=self.mark_after, code=self.code)
        def invert(self):
            self.data = self.data.inverted()
        def rotate_by(self, model, totalmove):
            rmove = model.rotate_move(totalmove.data, self.data)
            self.data = self.data.__class__(*rmove)
            self.code = None
        @classmethod
        def normalize_complete_moves(cls, model, totalmoves):
            for move in model.normalize_complete_moves(move.data for move in totalmoves):
                yield cls(MoveData(*move))
        def update_fullmoves(self, model):
            self.data = self.data.update_fullmoves(model)
            
    def __init__(self):
        self.current_place = self._pos_0
        self.moves = []
        
    def copy(self):
        movequeue = self.__class__()
        movequeue.current_place = self.current_place
        movequeue.moves = [m.copy() for m in self.moves]
        return movequeue
        
    def __str__(self):
        return '{}(len={})'.format(self.__class__.__name__, len(self.moves))
        
        
class MoveQueue (_MoveQueueBase):
    _pos_0 = 0
    
    @classmethod
    def new_from_code(cls, code, pos, model):
        moves = cls()
        mpos, cpos = moves.parse(code, pos, model)
        return moves, mpos, cpos
        
    def at_start(self):
        return self.current_place == 0
    def at_end(self):
        return self.current_place == self.queue_length
    @property
    def _prev(self):
        return self.moves[self.current_place-1]
    @property
    def _current(self):
        return self.moves[self.current_place]
    @property
    def queue_length(self):
        return len(self.moves)

    def push(self, move_data, **kwargs):
        new_element = self.MoveQueueItem(MoveData(*move_data), **kwargs)
        self.moves.append(new_element)

    def push_current(self, move_data):
        if not self.at_end():
            self.moves[self.current_place:] = []
        self.push(move_data)

    def current(self):
        return None if self.at_end() else self._current.data

    def retard(self):
        if not self.at_start():
            self.current_place -= 1
    
    def rewind_start(self):
        self.current_place = 0
        
    def forward_end(self):
        self.current_place = self.queue_length
        
    def truncate(self):
        self.moves[self.current_place:] = []
        
    def truncate_before(self):
        self.moves[:self.current_place] = []
        self.current_place = 0
        
    def reset(self):
        self.current_place = 0
        self.moves[:] = []
        
    def advance(self):
        if not self.at_end():
            self.current_place += 1
        
    def swapnext(self):
        cp = self.current_place
        if cp+1 < self.queue_length:
            ma, mb = self.moves[cp], self.moves[cp+1]
            ma.code = mb.code = None
            ma.data, mb.data = mb.data, ma.data
            self.advance()
            
    def swapprev(self):
        cp = self.current_place
        if 0 < cp < self.queue_length:
            ma, mb = self.moves[cp-1], self.moves[cp]
            ma.code = mb.code = None
            ma.data, mb.data = mb.data, ma.data
            self.retard()
            
    def invert(self):
        # a mark at the end of the moves is discarded because a mark at start is not supported
        mark = False
        for move in self.moves:
            move.invert()
            move.code = None
            move.mark_after, mark = mark, move.mark_after
        self.moves.reverse()
        if not (self.at_start() or self.at_end()):
            self.current_place = len(self.moves) - self.current_place
            
    def normalize_complete_rotations(self, model):
        totalmoves = []
        new_moves = []
        for i, move in enumerate(self.moves):
            if i == self.current_place:
                self.current_place = len(new_moves)
            if move.slice < 0:
                totalmoves.append(move)
            else:
                for totalmove in reversed(totalmoves):
                    move.rotate_by(model, totalmove)
                new_moves.append(move)
        totalmoves = list(self.MoveQueueItem.normalize_complete_moves(model, totalmoves))
        self.moves = new_moves + totalmoves
        
    def push_packed(self, pmove):
        for move in pmove.data.unpack():
            self.push(move)
        
    def normalize_moves(self, model):
        mqp = MoveQueuePacked()
        mqp_current_place = (0, 0)
        for i, move in enumerate(self.moves):
            if i == self.current_place:
                mqp_current_place = (len(mqp.moves)-1, mqp.moves[-1].data.length()) if mqp.moves else (0, 0)
            mqp.push_unpacked(move.data, model)
        mqp_moves = mqp.separate_full_moves(model)
        if len(self.moves) <= self.current_place:
            mqp_current_place = len(mqp_moves), 0
        self.reset()
        for i, move in enumerate(mqp_moves):
            if i == mqp_current_place[0]:
                self.current_place = len(self.moves) + mqp_current_place[1]
            self.push_packed(move)
        if len(mqp_moves) <= mqp_current_place[0]:
            self.current_place = len(self.moves)
        
    def is_mark_current(self):
        return self.at_start() or self._prev.mark_after
        
    def is_mark_after(self, pos):
        return self.moves[pos].mark_after
        
    def mark_current(self, mark=True):
        if not self.at_start():
            self._prev.mark_after = mark
            if self._prev.code is not None:
                self._prev.code = self._prev.code.replace(' ','')
                if mark:
                    self._prev.code += ' '
                    
    def mark_and_extend(self, other):
        if not other.moves:
            return -1
        self.mark_current()
        self.truncate()
        self.moves += other.moves
        return self.current_place + other.current_place
        
    def format(self, model):
        code = ''
        pos = 0
        for i, move in enumerate(self.moves):
            if move.code is None:
                move.code = MoveFormat.format(move, model)
            code += move.code
            if i < self.current_place:
                pos = len(code)
        return code, pos
        
    def parse_iter(self, code, pos, model):
        code = code.lstrip(' ')
        queue_pos = self.current_place
        move_code = ''
        for i, c in enumerate(code):
            if move_code and MoveFormat.isstart(c, model):
                data, mark = MoveFormat.parse(move_code, model)
                if data is not None:
                    #FIXME: invalid chars at start get lost, other invalid chars are just ignored
                    self.push(data, mark_after=mark, code=move_code)
                yield data, queue_pos, i
                if i == pos:
                    queue_pos = self.queue_length
                move_code = c
            else:
                move_code += c
            if i < pos:
                queue_pos = self.queue_length + 1
        if move_code:
            data, mark = MoveFormat.parse(move_code, model)
            if data is not None:
                self.push(data, mark_after=mark, code=move_code)
            if len(code)-len(move_code) < pos:
                queue_pos = self.queue_length
            yield data, queue_pos, len(code)
            
    def parse(self, code, pos, model):
        queue_pos = 0
        cpos = 0
        for unused_data, queue_pos, code_len in self.parse_iter(code, pos, model):
            if cpos < pos:
                cpos = code_len
        return queue_pos, cpos
        
        
class MoveQueuePacked (_MoveQueueBase):
    _pos_0 = (0, 0, 0)
    
    def length(self):
        i = 0
        for move in self.moves:
            i += move.data.length()
        return i
        
    def axis_length(self, axis_dirs, axis, model):
        symmetry = model.symmetries[axis]
        if model.ignored_slice[axis] is None:
            vals = [0] * symmetry
            for m_dirs in axis_dirs:
                for s_base in range(symmetry):
                    v = m_dirs + s_base
                    if v > symmetry//2:
                        v -= symmetry
                    vals[s_base] += abs(v)
            minval = vals[0]+1
            for i, v in enumerate(vals):
                if v < minval:
                    f = i
                    minval = v
            f = symmetry - f
            if f > symmetry//2:
                f -= symmetry
            return minval, f
        else:
            f = axis_dirs[model.ignored_slice[axis]]
            length = 0
            for m_dirs in axis_dirs:
                v = m_dirs - f + symmetry
                if v > symmetry//2:
                    v -= symmetry
                length += abs(v)
            return length, f
            
    def length_no_full(self, model):
        i = 0
        axis = None
        axis_dirs = None
        for move in self.moves:
            if axis != move.data.axis:
                if axis is not None:
                    i += self.axis_length(axis_dirs, axis, model)[0]
                axis = move.data.axis
                axis_dirs = [0] * model.sizes[axis]
            assert move.data.slice >= 0
            assert axis_dirs[move.data.slice] == 0
            axis_dirs[move.data.slice] = move.data.dirs
        if axis is not None:
            i += self.axis_length(axis_dirs, axis, model)[0]
        return i
        
    def _separate_full_moves(self, i_axis, j_axis, axis, fullmoves, model):
        if not fullmoves:
            return self.moves[i_axis:j_axis]
        mslice = 0
        newmoves = []
        for i in range(i_axis, j_axis):
            move = self.moves[i]
            for j in range(mslice, move.data.slice):
                moveitem = self.MoveQueueItem(MoveDataPacked(axis, j, -fullmoves, model))
                newmoves.append(moveitem)
            assert axis == move.data.axis
            move.data = move.data.add_dirs(-fullmoves, model)
            newmoves.append(move)
            mslice = move.data.slice + 1
        for j in range(mslice, model.sizes[axis]):
            moveitem = self.MoveQueueItem(MoveDataPacked(axis, j, -fullmoves, model))
            newmoves.append(moveitem)
        moveitem = self.MoveQueueItem(MoveDataPacked(axis, -1, fullmoves))
        newmoves.append(moveitem)
        return newmoves
        
    def separate_full_moves(self, model):
        i_axis = 0
        axis = None
        axis_dirs = None
        newmoves = []
        for i, move in enumerate(self.moves):
            if axis != move.data.axis:
                if axis is not None:
                    fullmoves = self.axis_length(axis_dirs, axis, model)[1]
                    newmoves += self._separate_full_moves(i_axis, i, axis, fullmoves, model)
                i_axis = i
                axis = move.data.axis
                axis_dirs = [0] * model.sizes[axis]
            assert move.data.slice >= 0
            assert axis_dirs[move.data.slice] == 0, (axis_dirs, i, move.data.slice)
            axis_dirs[move.data.slice] = move.data.dirs
        if axis is not None:
            fullmoves = self.axis_length(axis_dirs, axis, model)[1]
            newmoves += self._separate_full_moves(i_axis, len(self.moves), axis, fullmoves, model)
        return newmoves
        
    def push_unpacked(self, move, model):
        # this function assumes that self.moves does not contain full moves,
        if move.slice < 0:
            # resolve full moves to slice moves
            for i in range(model.sizes[move.axis]):
                self.push_unpacked(MoveData(move.axis, i, move.dir), model)
            return
            
        i = len(self.moves)
        while True:
            if i <= 0:
                moveitem = None
                break
            # get item
            _moveitem = self.moves[i-1]
            parallel = model.parallel(_moveitem.data, move)
            if parallel == parallel.NO or parallel == parallel.LT:
                moveitem = None
                break
            moveitem = _moveitem
            i -= 1
            if parallel == parallel.COMPAT:
                break
            assert parallel == parallel.GT
        # create new empty item if needed
        if moveitem is None:
            moveitem = self.MoveQueueItem(MoveDataPacked(move.axis, move.slice, 0))
            self.moves.insert(i, moveitem)
        # combine move with the compatible moveitem.data
        assert move.axis == moveitem.data.axis
        moveitem.data = moveitem.data.add_dirs((-1 if move.dir else 1), model)
        # remove if empty
        if moveitem.data.dirs == 0:
            self.moves.pop(i)
            
    def format(self, model):
        mq = MoveQueue()
        for move in self.moves:
            mq.push_packed(move)
        return mq.format(model)
        
                
class MoveFormat: # pylint: disable=W0232
    re_flubrd = re.compile(r"(.)(\d*)(['-]?)([^ ]*)( *)(.*)")
    
    @classmethod
    def isstart(cls, char, model):
        return char.upper() in model.faces
    
    @staticmethod
    def intern_to_str_move(move, model):
        if move.slice <= -1:
            # Rotate entire cube
            if move.dir:
                mface = model.symbolsI[move.axis]
                if mface in model.faces:
                    # opposite symbol not reversed
                    return mface, '', ''
                # no opposite symbol
            mface = model.symbols[move.axis]
            mdir = '-' if move.dir else ''
            return mface, '', mdir
        elif move.slice*2 > model.sizes[move.axis]-1:
            mface = model.symbolsI[move.axis]
            if mface in model.faces:
                # slice is nearer to the opposite face
                mface = mface.lower()
                mslice = model.sizes[move.axis]-1 - move.slice
                mslice = str(mslice+1) if mslice else ''
                mdir = '' if move.dir else '-'
                return mface, mslice, mdir
            # no opposite face
        mface = model.symbols[move.axis].lower()
        mslice = str(move.slice+1) if move.slice else ''
        mdir = '-' if move.dir else ''
        return mface, mslice, mdir
    
    @classmethod
    def format(cls, move, model):
        mface, mslice, mdir = cls.intern_to_str_move(move, model)
        mark = ' ' if move.mark_after else ''
        move_code = mface + mslice + mdir + mark
        return move_code
        
    @staticmethod
    def str_to_intern_move(tface, tslice, tdir, model):
        mface = tface.upper()
        mslice = int(tslice or 1) - 1
        mdir = bool(tdir)
        if mface not in model.faces:
            return None
        elif mface in model.symbols:
            maxis = model.symbols.index(mface)
        elif mface in model.symbolsI:
            maxis = model.symbolsI.index(mface)
            mslice = model.sizes[maxis]-1 - mslice
            mdir = not mdir
        else:
            assert False, 'faces is a subset of symbols+symbolsI'
        if mslice < 0 or mslice >= model.sizes[maxis]:
            return None
        if tface.isupper():
            mslice = -1
        return MoveData(maxis, mslice, mdir)
        
    @classmethod
    def parse(cls, move_code, model):
        tface, tslice, tdir, unused_err1, mark, unused_err2 = cls.re_flubrd.match(move_code).groups()
        mark = bool(mark)
        move = cls.str_to_intern_move(tface, tslice, tdir, model)
        if move is None:
            debug('Error parsing formula')
            return move, False
        return move, mark
        
        



#-*- coding:utf-8 -*-

#  Copyright © 2009-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import math
from collections import namedtuple, OrderedDict

from .debug import (debug, DEBUG, DEBUG_MSG, DEBUG_DRAW, DEBUG_ROTATE, DEBUG_ERROR, DEBUG_MOUSEPOS,
                    DEBUG_LIVEQML, DEBUG_LIVESHADER, DEBUG_LIVEPLUGINS, DEBUG_FUNC,
                    DEBUG_VFPS, DEBUG_EMPTY,
                   )
from . import config
from .settings import settings
from .theme import Theme
from .ext import glarea, qt
from . import pluginlib
from .model import Model, empty_model
from .gamestate import GameState
from .utils import ExceptionMeta

try:
    _, ngettext
except NameError:
    _ = lambda string: string
    ngettext = lambda singular, plural, cnt: singular if cnt == 1 else plural
    
if DEBUG_FUNC:
    from .debug import meta
    Meta = meta(ExceptionMeta)
else:
    Meta = ExceptionMeta
    
    
class AnimType: # pylint: disable=R0903
    __slots__ = ()
    NONE, FWD, BWD, KEY, NEW = range(5)
AnimType = AnimType()
    
    
class MainWindow (metaclass=Meta):
    ### init ###
    def __init__(self, opts):
        self.games_file = opts.games_file
        self.qmlfile = opts.qml_file
        
        # dialogs
        self.preferences_dialog_created = False
        self.progress_dialog = None  # created when needed
        
        # plugins
        self.plugin_helper = pluginlib.PluginHelper()
        self.challenge_open = False
        self.challenge = None
        
        # glarea
        self.theme = Theme()
        self.mirror_distance = None
        self.glpickdata = None
        self.pickdata = None
        self.use_cursor = False
        self.last_mouse_x = -1
        self.last_mouse_y = -1
        self.button_down_background = False
        self.editing_model = False
        self.drop_colorname = None
        self.drop_filename = None
        self.animation_active = False
        self.stop_requested = False
        
        self.animtype = AnimType.NONE
        self.gesture_total = None
        self.move_keys = {}
        self.model = empty_model
        self.game = GameState()
        
        Model.load_index()  # needed by settings
        
        # initialize settings
        version = settings.load(opts.config_file)
        if version == 1:
            from . import schema
            schema.migrate_theme_face_N_color_image_mode(settings)
            values = schema.migrate_game_size_blocks_moves_position(settings)
            #XXX: block indices are not the same in previous version
            #if values is not None:
            #    GameState.migrate_1to2(self.games_file, *values)
        if version < 3:
            from . import schema
            schema.migrate_2_3(settings)
            
        if opts.demo is not None:
            import importlib
            demo = importlib.import_module(opts.demo)
            demo.run(self)
            
    def cb_before_mainloop(self):
        qt.create_window(self)
        # UI
        fixedshaders = self.gl_read_shader('pick.vert'), self.gl_read_shader('pick.frag')
        if DEBUG_DRAW:
            fixedshaders += self.gl_read_shader('hud.vert'),
        self.theme_init()
        self.ui_update_cursor()
        
        glarea.init_engine()
        self.gl_update_background_color()
        self.gl_prepare_faces_atlas()
        glarea.set_fixedshaders(fixedshaders)
        self.gl_read_shaders()
        glarea.set_antialiasing(qt.get_samples() > 1)
        self.camrotation = glarea.set_rotation_xy(*self.model.default_rotation)
        self.last_camrotation = self.camrotation
        
        self.ui_toolbar_empty()
        qt.load_ui(self.qmlfile)
        self.ui_fill_sidepane()
        self.move_keys = self.get_move_key_dict()
        
        settings.connnect(self.on_settings_changed)
        
        qt.initial_resize()
        
        if DEBUG:
            self.init_live_hacking()
            
    def on_init_finalize(self):
        if DEBUG_EMPTY:
            self.ui_update()
            qt.update_sidepane()
        else:
            self.ui_load_current_game()
            
    def cb_after_mainloop(self):
        qt.destroy_window()
        
    def on_qml_directoryChanged(self):  # only variant qtq
        print('reloading qml')
        qt.load_ui(self.qmlfile)
        self.ui_update()
            
    def on_reload_plugins(self):
        import importlib
        for name, module in sys.modules.items():
            if name.startswith(('pybiklib.plugins.')):
                print('reloading', name)
                importlib.reload(module)
        self.ui_fill_sidepane()
        qt.update_sidepane()
        
    def init_live_hacking(self):
        if DEBUG_LIVEQML:  # only variant qtq
            qt.create_filesystemwatcher(os.path.join(config.UI_DIR, 'qt'), self.on_qml_directoryChanged)
            if os.path.dirname(self.qmlfile) != os.path.join(config.UI_DIR, 'qt'):
                qt.create_filesystemwatcher(self.qmlfile, self.on_qml_directoryChanged)
        if DEBUG_LIVESHADER:
            qt.create_filesystemwatcher(config.SHADER_DIR, self.on_gl_reload_shader)
        if DEBUG_LIVEPLUGINS:
            from . import plugins
            qt.create_filesystemwatcher(config.PLUGINS_DIR, self.on_reload_plugins)
            qt.create_filesystemwatcher(os.path.dirname(plugins.__file__), self.on_reload_plugins)
            
        
    ### helper functions ###
    
    def theme_load_face_image(self, facekey):
        filename = settings['theme.faces',facekey,'image']
        if not self.theme.load_face_image(facekey, filename):
            del settings['theme.faces',facekey,'image']
        
    def theme_init(self):
        for facekey, unused_facename in Model.cache_index['facenames']:
            color = qt.parse_color(settings['theme.faces',facekey,'color'])
            self.theme.faces[facekey].color = color
            self.theme.faces[facekey].imagemode = settings['theme.faces',facekey,'mode']
            self.theme_load_face_image(facekey)
            
    @staticmethod
    def get_move_key_dict():
        return {qt.keyval_from_name(key): move for (move, key) in settings['draw.accels']}
        
    def ui_fill_sidepane(self):
        # get a list of all plugins
        plugins = self.plugin_helper.load_plugins()
        # convert plugin list to a tree structure
        ptree = OrderedDict()
        for plugin_path, func in plugins:
            subtree = [None, ptree, None]
            for name, transl in plugin_path:
                subtree = subtree[1].setdefault(name, [None, OrderedDict(), transl])
            if plugin_path:
                subtree[0] = func
        def genplugintree(tree):
            for unused_name, [func, subtree, transl] in tree.items():
                yield transl, func, list(genplugintree(subtree))
        ptree = list(genplugintree(ptree))
        qt.fill_sidepane(ptree)
        
    def on_plugin_test_idx(self, funcidx):
        if funcidx is not None:
            try:
                self.plugin_helper.get_function(self.game.current_state.model, funcidx)
            except pluginlib.PluginModelCompatError:
                return False
        return True
        
    def ui_load_game(self, mtype, size):
        model = Model(mtype, size)
        self.game.load(self.games_file, model)
        self.model = model
        self.camrotation = glarea.set_rotation_xy(*self.model.default_rotation)
        self.gl_update_frustum()
        self.gl_update_gldata()
        self.gl_set_transformations(self.game.current_state.rotations)
        self.gl_update_selection()
        qt.update_drawingarea()
        qt.set_title(_(config.APPNAME) +' – '+ str(self.game.current_state.model))
        self.ui_update()
        qt.update_sidepane()
        return model.type, model.size
        
    def ui_load_current_game(self):
        mtype = settings['game.type']
        size = settings['games',mtype,'size']
        self.ui_load_game(mtype, size)
        
    def ui_load_other_game(self, mtype, size):
        self.ui_save_game()
        if self.animation_active:
            return
        assert mtype is not None
        self.challenge = None
        self.challenge_open = False
        mtype, size = self.ui_load_game(mtype, size)
        settings['game.type'] = mtype
        settings['games',mtype,'size'] = size
        
    on_load_other_game = ui_load_other_game
    
    def ui_save_game(self):
        self.game.save(self.games_file)
        
    def ui_new_game(self, solved=False):
        if self.animation_active:
            return
        self.game.random(count=0 if solved else -1)
        self.challenge_open = not solved
        self.gl_update_gldata()
        self.gl_set_transformations(self.game.current_state.rotations)
        self.gl_update_selection()
        qt.update_drawingarea()
        self.ui_update()
        
    def animation_start(self, move_data, stop_after, animtype):
        if not move_data:
            return False
        self.animation_active = True
        self.animtype = animtype
        self.stop_requested = stop_after
        
        blocks = self.game.current_state.identify_rotation_blocks(
                            move_data.axis, move_data.slice)
        angle_max = 360. / self.model.symmetries[move_data.axis]
        axisx, axisy, axisz = self.model.axes[move_data.axis]
        if move_data.dir:
            glarea.set_animation_start(blocks, -axisx, -axisy, -axisz)
        else:
            glarea.set_animation_start(blocks, axisx, axisy, axisz)
        qt.animate_timer_start(0 if DEBUG_VFPS else 20, angle_max)
        self.gl_update_selection()
        qt.update_drawingarea()
        self.ui_toolbar_playing(self.game.mark_before)
        return True
        
    def ui_update_statusbar(self):
        '''This function must be called before any action that change the move queue length
        to reflect total count of moves and after each single move'''
        
        if self.editing_model:
            mesg = _('Press the Esc key to exit Edit Mode')
            self.challenge_open = False
        else:
            current = self.game.move_sequence.current_place
            total = self.game.move_sequence.queue_length
            solved = self.game.current_state.is_solved()
            
            #TRANSLATORS: substitution for {move_text} in statusbar text
            move_text = ngettext("{current} / {total} move",
                                 "{current} / {total} moves",
                                 total).format(current=current, total=total)
            #TRANSLATORS: substitution for {solved_text} in statusbar text
            solved_text = _('solved') if solved else _('not solved')
            
            #TRANSLATORS: statusbar text
            mesg = _('{move_text}, {solved_text}').format(
                            move_text=move_text,
                            solved_text=solved_text)
            if self.challenge_open and solved:
                self.challenge_open = False
                qt.show_message(_('Congratulations, you have solved the puzzle!'))
        qt.set_status_text(mesg)
        
    def ui_update(self):
        # update toolbar
        if self.game.move_sequence.at_start():
            if self.game.move_sequence.at_end():
                self.ui_toolbar_empty()
            else:
                self.ui_toolbar_first()
        else:
            if self.game.move_sequence.at_end():
                self.ui_toolbar_last(self.game.move_sequence.is_mark_after(-1))
            else:
                self.ui_toolbar_mid(self.game.move_sequence.is_mark_current())
                
        # update formula
        code, pos = self.game.move_sequence.format(self.game.current_state.model)
        qt.set_edit_moves(code, pos)
        
        self.ui_update_statusbar()
        
    def ui_update_progress(self, step, message=None, value_max=None):
        if self.progress_dialog is None:
            from . import dialogs
            self.progress_dialog = dialogs.ProgressDialog()
        canceled = self.progress_dialog.tick(step, message, value_max)
        return canceled
        
    def ui_end_progress(self):
        if self.progress_dialog is not None:
            self.progress_dialog.done()
            
    ### glarea helpers ###
    
    def gl_prepare_faces_atlas(self):
        width, height = 0, 0
        faces = [self.theme.faces[fk] for fk, fn in Model.cache_index['facenames']]
        atlasdata = []
        for fd in faces:
            x = width
            w, h, data = fd.image
            width += w
            height = max(h, height)
            atlasdata.append((data, x, w, h))
        w = h = 64
        while w < width:
            w *= 2
        while h < height:
            h *= 2
        width, height = w, h
        
        for fd, (unused_data, x, w, h) in zip(faces, atlasdata):
            fd.texrect = (x+.5)/width, .5/height, (x+w-.5)/width, (h-.5)/height
        qt.set_atlas_size(width, height)
        for i, (data, x, w, h) in enumerate(atlasdata):
            glarea.set_atlas_data(i, data, x, w, h)
            
    @staticmethod
    def gl_update_background_color():
        glarea.set_background_color(*qt.parse_color_f(settings['theme.bgcolor']))
        
    @staticmethod
    def gl_update_samples():
        if qt.get_samples():
            samples = settings['draw.samples']
            samples = 2**samples
            if samples == 1:
                glarea.set_antialiasing(False)
            else:
                glarea.set_antialiasing(True)
            qt.update_drawingarea()
            
    @staticmethod
    def gl_update_zoom():
        zoom = settings['draw.zoom']
        glarea.set_frustum(0, zoom)
        qt.update_drawingarea()
        
    def gl_update_selection(self):
        if self.animation_active:
            if self.pickdata is not None:
                self.pickdata = None
                self.ui_update_cursor()
            return
        if self.use_cursor:
            qt.set_pick_requested(True)
            qt.update_drawingarea()
            
    def ui_update_cursor(self):
        if not self.use_cursor:
            return
        if self.pickdata is None or self.button_down_background:
            index = -1
        elif self.pickdata.angle is None:
            index = -2
        else:
            index = int((self.pickdata.angle+180) / 22.5 + 0.5) % 16
        qt.set_cursor(index)
        
    def ui_unset_cursor(self):
        if self.use_cursor:
            qt.unset_cursor()
            
    def gl_set_transformations(self, rotations):
        self.rotations = rotations
        glarea.set_transformations(rotations)
        
    def gl_update_gldata(self, editing=False):
        selection_mode = 0 if editing else settings['draw.selection']
        glmodeldata, self.glpickdata = self.model.gl_vertex_data(selection_mode, self.theme, self.mirror_distance)
        glarea.set_data(*glmodeldata)
        use_cursor = selection_mode != 2
        if use_cursor != self.use_cursor:
            if use_cursor:
                self.use_cursor = use_cursor
                self.ui_update_cursor()
            else:
                self.ui_unset_cursor()
                self.use_cursor = use_cursor
                
    def gl_update_frustum(self):
        #TODO: The bounding_sphere_radius is optimised for the far clipping plane,
        #      for the near clipping plane the radius without the mirror_distance
        #      would be sufficient.
        s = max(self.model.sizes) if self.model.sizes else 1
        if settings['draw.mirror_faces']:
            self.mirror_distance = settings['draw.mirror_distance'] * s
            sm = s + self.mirror_distance
        else:
            self.mirror_distance = None
            sm = s
        bounding_sphere_radius = math.sqrt(2*s*s + sm*sm)
        glarea.set_frustum(bounding_sphere_radius, settings['draw.zoom'])
        
    def ui_set_editing_mode(self, enable):
        self.editing_model = enable
        self.gl_update_gldata(editing=enable)
        self.gl_update_selection()
        
    @staticmethod
    def gl_read_shader(filename):
        if DEBUG_MSG or DEBUG_LIVESHADER:
            print('Loading shader:', filename)
        filename = os.path.join(config.SHADER_DIR, filename)
        with open(filename, 'rb') as sfile:
            source = sfile.read()
        return source
        
    @classmethod
    def gl_read_shaders(self):
        shadername = settings['draw.shader_nick']
        shaders = self.gl_read_shader(shadername+'.vert'), self.gl_read_shader(shadername+'.frag')
        glarea.set_shaders(*shaders)
        
    def gl_reload_shader(self):
        self.gl_read_shaders()
        qt.update_drawingarea()
    on_gl_reload_shader = gl_reload_shader
        
    def ui_update_dropped_color(self):
        if self.pickdata is not None:
            pickdata = self.pickdata
            if type(pickdata) is list:
                pickdata = pickdata[0]
            unused_idx, unused_rotnum, rot = self.rotations[pickdata.blockpos]
            colorsym = self.model.face_symbolic_to_face_color(pickdata.symbol, rot)
            colornum = self.model.faces.index(colorsym)
            facekey = self.model.facekeys[colornum]
            settings['theme.faces',facekey,'color'] = self.drop_colorname
        else:
            settings['theme.bgcolor'] = self.drop_colorname
            
    def ui_update_dropped_file(self):
        if self.pickdata is not None:
            pickdata = self.pickdata
            if type(pickdata) is list:
                pickdata = pickdata[0]
            unused_idx, unused_rotnum, rot = self.rotations[pickdata.blockpos]
            colorsym = self.model.face_symbolic_to_face_color(pickdata.symbol, rot)
            colornum = self.model.faces.index(colorsym)
            facekey = self.model.facekeys[colornum]
            settings['theme.faces',facekey,'image'] = self.drop_filename
        else:
            debug('Background image is not supported.')
            
    ### application handlers ###
    
    def on_closing(self):
        settings.disconnect()
        self.ui_save_game()
        settings.flush()
        
    def on_settings_timer_timeout(self):
        settings.flush()
        
    def on_settings_changed(self, key):
        qt.settings_timer_start()
        if key is None:
            return
        if key == 'draw.accels':
            self.move_keys = self.get_move_key_dict()
        elif key == 'draw.speed':
            qt.set_animation_speed()
        elif key == 'draw.shader':
            self.gl_reload_shader()
        elif key == 'draw.samples':
            self.gl_update_samples()
        elif key == 'draw.selection':
            self.gl_update_gldata()
            self.gl_update_selection()
            qt.update_drawingarea()
        elif key in ['draw.mirror_faces', 'draw.mirror_distance']:
            self.gl_update_frustum()
            self.gl_update_gldata()
            self.gl_update_selection()
            qt.update_drawingarea()
        elif key.startswith('theme.faces.'):
            facekey, faceattr = key.split('.')[2:]
            if faceattr == 'color':
                color = qt.parse_color(settings['theme.faces',facekey,'color'])
                self.theme.faces[facekey].color = color
                self.gl_update_gldata()
            elif faceattr == 'image':
                self.theme_load_face_image(facekey)
                self.gl_prepare_faces_atlas()
                self.gl_update_gldata()
            elif faceattr == 'mode':
                self.theme.faces[facekey].imagemode = settings['theme.faces',facekey,'mode']
                self.gl_update_gldata()
            qt.update_drawingarea()
        elif key == 'theme.bgcolor':
            self.gl_update_background_color()
            qt.update_drawingarea()
        elif key == 'draw.zoom':
            self.gl_update_zoom()
        else:
            debug('Unknown settings key changed:', key)
            
    def on_key_pressed(self, key, modifiers, fromdrawingarea):
        if DEBUG_ERROR and key == qt.QtKeys.Key_Period:
            print('DEBUG_ERROR')
            qt.show_message('DEBUG_ERROR from on_key_pressed')
            return True
            
        if fromdrawingarea:
            if modifiers & (qt.QtKeys.ShiftModifier | qt.QtKeys.ControlModifier):
                pass
            elif key == qt.QtKeys.Key_Right:
                self.camrotation_changed(2, 0)
                return True
            elif key == qt.QtKeys.Key_Left:
                self.camrotation_changed(-2, 0)
                return True
            elif key == qt.QtKeys.Key_Up:
                self.camrotation_changed(0, -2)
                return True
            elif key == qt.QtKeys.Key_Down:
                self.camrotation_changed(0, 2)
                return True
                
        if key == qt.QtKeys.Key_Escape:
            self.ui_set_editing_mode(False)
            self.ui_update()
            return True
        try:
            move = self.move_keys[key | modifiers]
        except KeyError:
            return False
        else:
            if self.animation_active and self.animtype != AnimType.KEY:
                return True
            if not self.animation_active:
                self.game.move_sequence.truncate()
            self.game.move_sequence.parse(move, len(move), self.game.current_state.model)
            self.ui_update()
            if not self.animation_active:
                move_data = self.game.step_next()
                self.animation_start(move_data, stop_after=False, animtype=AnimType.KEY)
        return True
            
    ### GL area handlers ###
    
    def animation_stop(self):
        self.animation_active = False
        self.stop_requested = False
        qt.animate_timer_stop()
        
    def animation_ending(self):
        self.gl_set_transformations(self.game.current_state.rotations)
        
        if not self.stop_requested:
            # go again
            self.ui_update_statusbar()
            if self.animtype == AnimType.BWD:
                move_data = self.game.step_back()
            else:
                move_data = self.game.step_next()
            if self.animation_start(move_data, stop_after=False, animtype=self.animtype):
                self.stop_requested = False
                self.gl_update_selection()
                qt.update_drawingarea()
                return
        self.animation_stop()
        self.ui_update()
        self.gl_update_selection()
        qt.update_drawingarea()
            
    on_animation_ending = animation_ending
    
    PickData = namedtuple('PickData', 'maxis mslice mdir blockpos symbol angle')
    
    def gl_resolve_pickindex(self, index):
        if not index:
            return None
        try:
            maxis, mslice, mdir, blockpos, symbol, arrowdir = self.glpickdata[index]
        except IndexError:
            #FIXME: should not happen, but it does
            return None
        
        if maxis is None:
            pickdata = []
            for maxis, mslice, arrowdir in mslice:
                angle = glarea.get_cursor_angle(arrowdir)
                pickdata.append(self.PickData(maxis, mslice, mdir, blockpos, symbol, angle))
            return pickdata
        if arrowdir is None:
            angle = None
        else:
            angle = glarea.get_cursor_angle(arrowdir)
        pickdata = self.PickData(maxis, mslice, mdir, blockpos, symbol, angle)
        return pickdata
        
    def debug_update_pick_text(self):
        if self.pickdata is not None:
            pos = self.pickdata.blockpos
            face = self.model.faces.index(self.pickdata.symbol)
            idx, unused_rotnum, rot = self.rotations[self.pickdata.blockpos]
            colorsym = self.model.face_symbolic_to_face_color(self.pickdata.symbol, rot)
            angle = self.pickdata.angle and format(self.pickdata.angle, '.1f')
            text = ('idx {idx}, pos {pos} {indices}, rot {rot!r}\n'
                    'face {0.symbol} ({face}), color {colorsym}\n'
                    'axis {0.maxis}, slice {0.mslice}, dir {0.mdir}, angle {angle}'
                   ).format(self.pickdata,
                    idx=idx, pos=pos, indices=self.model.cell_indices[pos], rot=rot,
                    colorsym=colorsym, angle=angle, face=face)
            qt.set_debug_text(text)
            
    def mouse_rotate_picked_slice(self, total, reverse):
        pd = self.pickdata
        move = pd and (pd.maxis,  (-1 if total else pd.mslice),  (not pd.mdir if reverse else pd.mdir))
        self.game.set_next(move)
        self.ui_update_statusbar()
        move_data = self.game.step_next()
        self.animation_start(move_data, stop_after=False, animtype=AnimType.NEW)
        
    def mouse_start_gesture(self, x, y, height):
        self.last_mouse_x = x
        self.last_mouse_y = y
        self.last_camrotation = self.camrotation
        glarea.set_pick_position(x, height-y)
        qt.set_pick_requested(True)
        qt.update_drawingarea()
        
    def mouse_get_gesture_move(self, x, y, total):
        pds = self.pickdata
        if not pds:
            return None
        dx = x - self.last_mouse_x
        dy = y - self.last_mouse_y
        if dx == dy == 0:
            return None
        angle = (math.atan2(-dx, dy) * 180. / math.pi) % 360.
        mangle = 360. * 2
        move = None
        for pd in pds:
            tangle = (20. if pd.mslice == -1 else 0.)
            dangle = min((angle-pd.angle)%360., (pd.angle-angle)%360.) + tangle
            if dangle < mangle:
                mangle = dangle
                move = (pd.maxis,  (-1 if total else pd.mslice),  pd.mdir)
            dangle = min((angle-pd.angle+180)%360., (pd.angle-angle+180)%360.) + tangle
            if dangle < mangle:
                mangle = dangle
                move = (pd.maxis,  (-1 if total else pd.mslice),  not pd.mdir)
        return move
        
    def mouse_rotate_gesture_slice(self, x, y):
        if self.gesture_total is None:
            return None
        move = self.mouse_get_gesture_move(x, y, self.gesture_total)
        self.gesture_total = None
        if move is None:
            return
        self.game.set_next(move)
        self.ui_update_statusbar()
        move_data = self.game.step_next()
        self.animation_start(move_data, stop_after=False, animtype=AnimType.NEW)
        
    def mouse_swap_picked_block(self):
        pd = self.pickdata
        self.game.swap_block(pd.blockpos, pd.maxis, pd.mslice, pd.mdir)
        self.gl_set_transformations(self.game.current_state.rotations)
        
    def mouse_rotate_picked_block(self, rdir):
        blockpos = self.pickdata.blockpos
        self.game.rotate_block(blockpos, rdir)
        self.gl_set_transformations(self.game.current_state.rotations)
        
    def camrotation_start(self, x=None, y=None):
        self.button_down_background = True
        if x is not None:
            self.last_mouse_x = x
            self.last_mouse_y = y
            self.last_camrotation = self.camrotation
            
    def camrotation_changed(self, dx, dy):
        rx, ry = self.camrotation
        self.camrotation = glarea.set_rotation_xy(rx+dx, ry+dy)
        #TODO: configurable rotation methods
        #glarea.set_rotation_rel(dx, dy)
        self.gl_update_selection()
        qt.update_drawingarea()
        
    def camrotation_end(self):
        self.button_down_background = False
        self.gl_update_selection()
            
    def on_mouse_pressed(self, modifiers, button, x, y, height):
        if DEBUG_MOUSEPOS:
            print('mouse press modifiers={} button={} x={} y={}'.format(modifiers, button, x, y))
        total = modifiers & qt.QtKeys.ControlModifier
        if settings['draw.selection'] == 2:
            self.gesture_total = total
            self.mouse_start_gesture(x, y, height)
        elif self.pickdata is not None:
            if self.animation_active:
                return
            # make a move
            if self.editing_model:
                if modifiers & qt.QtKeys.ControlModifier:
                    if button == qt.QtKeys.LeftButton:
                        self.mouse_rotate_picked_block(False)
                    elif button == qt.QtKeys.RightButton:
                        self.mouse_rotate_picked_block(True)
                else:
                    if button == qt.QtKeys.LeftButton:
                        self.mouse_swap_picked_block()
            else:
                if button == qt.QtKeys.LeftButton:
                    self.mouse_rotate_picked_slice(total, False)
                elif button == qt.QtKeys.RightButton and settings['draw.selection_nick'] == 'simple':
                    self.mouse_rotate_picked_slice(total, True)
        elif button == qt.QtKeys.LeftButton:
            self.camrotation_start(x, y)
        qt.update_drawingarea()
        
    def on_mouse_released(self, x, y):
        if DEBUG_MOUSEPOS:
            print('mouse release x={} y={}'.format(x, y))
        if self.button_down_background:
            self.camrotation_end()
        else:
            self.mouse_rotate_gesture_slice(x, y)
        
    def on_camrotation_changed(self, dx, dy):  # only variant qtq
        self.camrotation_changed(dx, dy)
        
    def on_mouse_entered(self):  # only variant qtq
        self.ui_update_cursor()
        
    def on_mouse_moved(self, buttons, x, y, height, pixeldensity):
        if self.use_cursor:
            glarea.set_pick_position(x, height-y)
        if DEBUG_MOUSEPOS:
            print('mouse {} buttons={} x={} y={}'.format('move' if buttons else 'hover', buttons, x, height-y))
        
        if not (buttons and self.button_down_background):
            self.gl_update_selection()
            return
            
        # perform rotation
        rx, ry = self.last_camrotation
        dx = (x - self.last_mouse_x) / pixeldensity * 4
        dy = (y - self.last_mouse_y) / pixeldensity * 4
        self.camrotation = glarea.set_rotation_xy(rx+dx, ry+dy)
        qt.update_drawingarea()
        
    def on_mouse_drop_color(self, x, height_y, color):
        if DEBUG_MOUSEPOS:
            print('dropped color', x, height_y, color)
        glarea.set_pick_position(x, height_y)
        self.drop_colorname = color
        qt.set_pick_requested(True)
        qt.update_drawingarea()
        
    def on_mouse_drop_url(self, x, height_y, url):
        if DEBUG_MOUSEPOS:
            print('dropped url', x, height_y, url)
        glarea.set_pick_position(x, height_y)
        import urllib.parse, urllib.request
        type, filename = urllib.parse.splittype(url)
        host, filename = urllib.parse.splithost(filename)
        if type != 'file' or host:
            debug('"%s" not a local file.' % url)
            return
        filename = urllib.request.url2pathname(filename)
        if not filename or not os.path.exists(filename):
            debug('"%s" not found.' % filename)
            return
        self.drop_filename = filename
        qt.set_pick_requested(True)
        qt.update_drawingarea()
        
    def on_mouse_zoom(self, delta):
        if DEBUG_MOUSEPOS:
            print('wheel event', delta)
        zoom = settings['draw.zoom'] * math.pow(1.1, -delta)
        zoom_min, zoom_max = settings['draw.zoom_range']
        if zoom < zoom_min:
            zoom = zoom_min
        if zoom > zoom_max:
            zoom = zoom_max
        settings['draw.zoom'] = zoom
        self.gl_update_selection()
            
    def on_picking_result(self, index):
        if self.animation_active:
            if self.pickdata is not None:
                self.pickdata = None
                self.ui_update_cursor()
            return
            
        self.pickdata = self.gl_resolve_pickindex(index)
        self.ui_update_cursor()
        if DEBUG_DRAW:
            self.debug_update_pick_text()
            qt.update_drawingarea()
        if self.drop_colorname is not None:
            self.ui_update_dropped_color()
            self.drop_colorname = None
        elif self.drop_filename is not None:
            self.ui_update_dropped_file()
            self.drop_filename = None
        elif not self.use_cursor and self.pickdata is None:
            self.camrotation_start()
        qt.set_pick_requested(False)
        
    ### sidepane handlers ###
    
    def ui_plugin_activated(self, func_idx):
        if self.animation_active:
            return
        if func_idx < 0:
            return
        if self.game.current_state.model is empty_model:
            return
        try:
            func = self.plugin_helper.get_function(self.game.current_state.model, func_idx)
        except pluginlib.PluginModelCompatError as e:
            qt.show_message(e.args[0])
            return
        # model accepted, now run the plugin function
        game = self.game.copy()
        try:
            func(game)
        except pluginlib.PluginSolverAbort as e:
            self.challenge = None
            qt.show_message(e.args[0])
            return
        if game.plugin_mode == 'append':
            # append moves
            self.challenge_open = False
            position = self.game.move_sequence.mark_and_extend(game.move_sequence)
            changed = position >= 0
            if changed:
                self.game.goto_next_pos(position)
                changed = 'animate'
        elif game.plugin_mode == 'replace':
            # replace game
            self.game.initial_state = game.initial_state
            self.game.current_state = game.current_state
            self.game.move_sequence = game.move_sequence
            self.challenge_open = False
            changed = True
        elif game.plugin_mode == 'challenge':
            # challenge
            self.game.initial_state = game.initial_state
            self.game.current_state = game.initial_state.copy()
            self.game.move_sequence.reset()
            self.challenge = func_idx
            self.challenge_open = True
            changed = True
        else:
            assert False
        if changed:
            self.gl_set_transformations(self.game.current_state.rotations)
            self.ui_update()
            if changed == 'animate':
                move_data = self.game.step_next()
                self.animation_start(move_data, stop_after=False, animtype=AnimType.NEW)
        qt.update_drawingarea()
        
    on_plugin_activated = ui_plugin_activated
    
    ### action handlers ###
    
    def on_action_challenge_triggered(self):
        if self.challenge is None:
            self.ui_new_game(solved=False)
        else:
            self.ui_plugin_activated(self.challenge)
    def on_action_new_solved_triggered(self):
        self.ui_new_game(solved=True)
    def on_action_selectmodel_triggered(self):
        from . import dialogs
        return dialogs.ModelSelection()
    def on_action_quit_triggered(self):  # only variant qtq
        qt.close_mainwindow()
    def on_action_preferences_triggered(self):
        if not self.preferences_dialog_created:
            from . import dialogs
            sample_buffers = max(qt.get_samples(), 1)
            dialogs.init_preferences(sample_buffers)
            self.preferences_dialog_created = True
        qt.show_preferences()
    def on_action_reset_rotation_triggered(self):
        '''Reset cube rotation'''
        self.camrotation = glarea.set_rotation_xy(*self.model.default_rotation)
        qt.update_drawingarea()
        
    def on_action_rewind_triggered(self):
        if self.animation_active:
            self.stop_requested = True
            if self.animtype == AnimType.BWD:
                self.game.step_next() # undo the already applied move
            self.game.goto_prev_mark()
            self.animation_ending()
        else:
            needs_update = self.game.goto_prev_mark()
            self.gl_set_transformations(self.game.current_state.rotations)
            if needs_update:
                qt.update_drawingarea()
                self.ui_update()
    def on_action_previous_triggered(self):
        if self.animation_active:
            if self.animtype == AnimType.BWD:
                # request another BWD move
                self.stop_requested = False
            else:
                self.stop_requested = True
                self.game.step_back()
            self.animation_ending()
            self.stop_requested = True
        else:
            move_data = self.game.step_back()
            self.animation_start(move_data, stop_after=True, animtype=AnimType.BWD)
    def on_action_stop_triggered(self):
        if self.animation_active:
            if self.stop_requested:
                self.animation_ending()
            else:
                self.stop_requested = True
    def on_action_play_triggered(self):
        if not self.animation_active:
            move_data = self.game.step_next()
            self.animation_start(move_data, stop_after=False, animtype=AnimType.FWD)
    def on_action_next_triggered(self):
        if self.animation_active:
            sr = self.stop_requested
            if self.animtype == AnimType.BWD:
                self.stop_requested = True
                self.game.step_next()
            else:
                self.stop_requested = False
            self.animation_ending()
            self.stop_requested = sr
        else:
            move_data = self.game.step_next()
            self.animation_start(move_data, stop_after=True, animtype=AnimType.FWD)
    def on_action_forward_triggered(self):
        if self.animation_active:
            self.stop_requested = True
            if self.animtype != AnimType.BWD:
                self.game.step_back() # undo the already applied move
            self.game.goto_next_mark()
            self.animation_ending()
        else:
            self.game.goto_next_mark()
            self.gl_set_transformations(self.game.current_state.rotations)
            qt.update_drawingarea()
            self.ui_update()
    def on_action_mark_set_triggered(self):
        self.game.move_sequence.mark_current(True)
        self.ui_update()
    def on_action_mark_remove_triggered(self):
        self.game.move_sequence.mark_current(False)
        self.ui_update()
        
    def on_action_initial_state_triggered(self):
        if self.animation_active:
            return
        self.game.set_as_initial_state()
        self.ui_update()
        
    ### edit formula handlers ###
    
    def on_edit_finished(self, code, pos):
        if self.animation_active:
            self.animation_stop()
        self.game.set_code(code, pos)
        self.gl_set_transformations(self.game.current_state.rotations)
        qt.update_drawingarea()
        self.ui_update()
        
    def on_edit_moves_nextword(self, code, pos):  # only variant qtw
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if cpos <= pos:
            moves.advance()
        unused_code, pos = moves.format(self.game.current_state.model)
        return pos
        
    def on_edit_moves_prevword(self, code, pos):  # only variant qtw
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if cpos >= pos:
            moves.retard()
        unused_code, pos = moves.format(self.game.current_state.model)
        return pos
        
    def on_edit_moves_swapnext(self, code, pos):  # only variant qtw
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if pos < cpos:
            moves.retard()
        moves.swapnext()
        code, pos = moves.format(self.game.current_state.model)
        qt.set_edit_moves(code, pos)
        
    def on_edit_moves_swapprev(self, code, pos):  # only variant qtw
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if pos < cpos:
            moves.retard()
        moves.swapprev()
        code, pos = moves.format(self.game.current_state.model)
        qt.set_edit_moves(code, pos)
        
    ### misc ###
    
    def on_action_edit_cube_triggered(self):
        self.animation_stop()
        self.game.goto_start()
        if DEBUG_ROTATE:
            self.game.current_state.debug_blocksymbols(allsyms=True)
        self.gl_set_transformations(self.game.current_state.rotations)
        self.ui_set_editing_mode(True)
        self.ui_update()
        
    def on_dialog_text(self, key):
        from . import dialogs
        return dialogs.DialogValues.get_text(key)
        
    def dialog_get_image_data(self, imagefile):  # only variant qtw
        stockfiles = Theme.textures.get_icons()
        index_icon = 1 + len(stockfiles)
        if imagefile.startswith('/'):
            index = index_icon
        else:
            try:
                index = stockfiles.index(imagefile) + 1
            except ValueError:
                index = 0
            imagefile = None
        return index_icon, imagefile, index
        
    def on_dialog_change_current_face(self, current_facekey):  # only variant qtw
        color = settings['theme.faces',current_facekey,'color']
        imageindex_icon, imagefile, imageindex = self.dialog_get_image_data(settings['theme.faces',current_facekey,'image'])
        imagemode = settings['theme.faces',current_facekey,'mode']
        qt.set_preferences_current_face_theme(color, imageindex_icon, imagefile, imageindex, imagemode)
        
    def on_dialog_change_current_image(self, current_facekey, index):  # only variant qtw
        stockfiles = Theme.textures.get_icons()
        if index == 0:
            filename = ''
        elif 0 < index <= len(stockfiles):
            filename = stockfiles[index-1]
        else:
            filename = qt.get_filedialog_imagefile()
            if filename == '':
                # canceled, set the old image
                filename = settings['theme.faces',current_facekey,'image']
                index_icon, imagefile, index = self.dialog_get_image_data(filename)
                qt.set_combobox_current_image(index_icon, imagefile, index)
            else:
                index_icon, imagefile, index = self.dialog_get_image_data(filename)
                qt.set_combobox_current_image(index_icon, imagefile, -1)
        settings['theme.faces',current_facekey,'image'] = filename
        
    def on_dialog_reset_current_image(self, current_facekey):  # only variant qtw
        del settings['theme.faces',current_facekey,'image']
        index_icon, imagefile, index = self.dialog_get_image_data(settings['theme.faces',current_facekey,'image'])
        qt.set_combobox_current_image(index_icon, imagefile, index)
        
    # a: rewind enabled, previous enabled
    # b: play enabled, next enabled, forward enabled
    # c: add_mark enabled, remove_mark enabled
    # d: stop visible, play !visible
    # e: add_mark visible, remove_mark !visible
    #                                                      a  b  c  d         e
    def ui_toolbar_empty(self):         qt.set_toolbar_state((0, 0, 0, 0,        0))
    def ui_toolbar_first(self):         qt.set_toolbar_state((0, 1, 0, 0,        0))
    def ui_toolbar_mid(self, mark):     qt.set_toolbar_state((1, 1, 1, 0, not mark))
    def ui_toolbar_last(self, mark):    qt.set_toolbar_state((1, 0, 1, 0, not mark))
    def ui_toolbar_playing(self, mark): qt.set_toolbar_state((1, 1, 0, 1, not mark))
    
    




# generated with: tools/create-gl-pxd.py gl pybiklib/ext gl_ogl.pxd glarea.py gldraw.py gl_ogl.pxd

from libc.stddef cimport ptrdiff_t
from libc.stdint cimport int32_t, intptr_t, int8_t, uint8_t


# defines from /usr/include/GL/gl.h:

cdef extern from 'GL/gl.h':
    enum: GL_FALSE
    enum: GL_TRUE
    enum: GL_UNSIGNED_BYTE
    enum: GL_FLOAT
    enum: GL_POINTS
    enum: GL_LINES
    enum: GL_TRIANGLES
    enum: GL_CCW
    enum: GL_BACK
    enum: GL_CULL_FACE
    enum: GL_DEPTH_TEST
    enum: GL_RGB
    enum: GL_RGBA
    enum: GL_MAX_TEXTURE_SIZE
    enum: GL_TEXTURE_2D
    enum: GL_VENDOR
    enum: GL_RENDERER
    enum: GL_VERSION
    enum: GL_EXTENSIONS
    enum: GL_INVALID_ENUM
    enum: GL_INVALID_VALUE
    enum: GL_INVALID_OPERATION
    enum: GL_OUT_OF_MEMORY
    enum: GL_DEPTH_BUFFER_BIT
    enum: GL_COLOR_BUFFER_BIT
    enum: GL_MULTISAMPLE
    enum: GL_SAMPLE_ALPHA_TO_COVERAGE
    enum: GL_SAMPLE_COVERAGE
    enum: GL_SAMPLE_BUFFERS
    enum: GL_SAMPLES
    enum: GL_SAMPLE_COVERAGE_VALUE
    enum: GL_SAMPLE_COVERAGE_INVERT


# defines from /usr/include/GL/glext.h:

cdef extern from 'GL/glext.h':
    enum: GL_ARRAY_BUFFER
    enum: GL_STATIC_DRAW
    enum: GL_MAX_VERTEX_ATTRIBS
    enum: GL_FRAGMENT_SHADER
    enum: GL_VERTEX_SHADER
    enum: GL_DELETE_STATUS
    enum: GL_COMPILE_STATUS
    enum: GL_LINK_STATUS
    enum: GL_VALIDATE_STATUS
    enum: GL_INFO_LOG_LENGTH
    enum: GL_ATTACHED_SHADERS
    enum: GL_ACTIVE_UNIFORMS
    enum: GL_ACTIVE_UNIFORM_MAX_LENGTH
    enum: GL_ACTIVE_ATTRIBUTES
    enum: GL_ACTIVE_ATTRIBUTE_MAX_LENGTH
    enum: GL_SHADING_LANGUAGE_VERSION


# typedefs from /usr/include/GL/gl.h:

cdef extern from *:
    ctypedef unsigned int	GLenum
    ctypedef unsigned char	GLboolean
    ctypedef unsigned int	GLbitfield
    ctypedef void		GLvoid
    ctypedef int		GLint
    ctypedef unsigned char	GLubyte
    ctypedef unsigned int	GLuint
    ctypedef int		GLsizei
    ctypedef float		GLfloat
    ctypedef float		GLclampf


# typedefs from /usr/include/GL/glext.h:

cdef extern from *:
    ctypedef ptrdiff_t GLsizeiptr
    ctypedef ptrdiff_t GLintptr
    ctypedef char GLchar


# other typedefs:

cdef extern from *:
    ctypedef GLchar* const_GLchar_ptr "const GLchar*"


# functions from /usr/include/GL/gl.h:

cdef extern from 'GL/gl.h':
    cdef void  glClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha ) nogil
    cdef void  glClear( GLbitfield mask ) nogil
    cdef void  glCullFace( GLenum mode ) nogil
    cdef void  glFrontFace( GLenum mode ) nogil
    cdef void  glEnable( GLenum cap ) nogil
    cdef void  glDisable( GLenum cap ) nogil
    cdef GLboolean  glIsEnabled( GLenum cap ) nogil
    cdef void  glGetBooleanv( GLenum pname, GLboolean *params ) nogil
    cdef void  glGetFloatv( GLenum pname, GLfloat *params ) nogil
    cdef void  glGetIntegerv( GLenum pname, GLint *params ) nogil
    cdef GLenum  glGetError() nogil
    cdef GLubyte *  glGetString( GLenum name ) nogil
    cdef void  glViewport( GLint x, GLint y,
                                    GLsizei width, GLsizei height ) nogil
    cdef void  glDrawArrays( GLenum mode, GLint first, GLsizei count ) nogil
    cdef void  glReadPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil
    cdef void  glTexImage2D( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLsizei height,
                                    GLint border, GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil
    cdef void  glTexSubImage2D( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                       GLvoid *pixels ) nogil
    cdef void  glActiveTexture( GLenum texture ) nogil


# functions from /usr/include/GL/glext.h:

cdef extern from 'GL/glext.h':
    cdef void  glBindBuffer (GLenum target, GLuint buffer) nogil
    cdef void  glDeleteBuffers (GLsizei n, GLuint *buffers) nogil
    cdef void  glGenBuffers (GLsizei n, GLuint *buffers) nogil
    cdef void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage) nogil
    cdef void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data) nogil
    cdef void  glAttachShader (GLuint program, GLuint shader) nogil
    cdef void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name) nogil
    cdef void  glCompileShader (GLuint shader) nogil
    cdef GLuint  glCreateProgram () nogil
    cdef GLuint  glCreateShader (GLenum type) nogil
    cdef void  glDeleteProgram (GLuint program) nogil
    cdef void  glDeleteShader (GLuint shader) nogil
    cdef void  glDetachShader (GLuint program, GLuint shader) nogil
    cdef void  glDisableVertexAttribArray (GLuint index) nogil
    cdef void  glEnableVertexAttribArray (GLuint index) nogil
    cdef void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil
    cdef void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil
    cdef GLint  glGetAttribLocation (GLuint program, GLchar *name) nogil
    cdef void  glGetProgramiv (GLuint program, GLenum pname, GLint *params) nogil
    cdef void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil
    cdef void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params) nogil
    cdef void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil
    cdef GLint  glGetUniformLocation (GLuint program, GLchar *name) nogil
    cdef void  glLinkProgram (GLuint program) nogil
    cdef void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length) nogil
    cdef void  glUseProgram (GLuint program) nogil
    cdef void  glUniform1i (GLint location, GLint v0) nogil
    cdef void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value) nogil
    cdef void  glValidateProgram (GLuint program) nogil
    cdef void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer) nogil

# GL version 2.0 needed

